"""
An ASGI middleware.

Based on Tom Christie's `sentry-asgi <https://github.com/encode/sentry-asgi>`.
"""

import inspect
import sys
from copy import deepcopy
from functools import partial
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.api import continue_trace
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations._asgi_common import (
    _get_headers,
    _get_ip,
    _get_path,
    _get_request_attributes,
    _get_request_data,
    _get_url,
    _RootPathInPath,
)
from sentry_sdk.integrations._wsgi_common import (
    DEFAULT_HTTP_METHODS_TO_CAPTURE,
)
from sentry_sdk.scope import Scope, should_send_default_pii
from sentry_sdk.sessions import track_session
from sentry_sdk.traces import (
    SOURCE_FOR_STYLE as SEGMENT_SOURCE_FOR_STYLE,
)
from sentry_sdk.traces import (
    SegmentNameSource,
    StreamedSpan,
)
from sentry_sdk.tracing import (
    SOURCE_FOR_STYLE,
    Transaction,
    TransactionSource,
)
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    CONTEXTVARS_ERROR_MESSAGE,
    HAS_REAL_CONTEXTVARS,
    ContextVar,
    _get_installed_modules,
    capture_internal_exceptions,
    event_from_exception,
    logger,
    nullcontext,
    qualname_from_function,
    reraise,
    transaction_from_function,
)

if TYPE_CHECKING:
    from typing import Any, ContextManager, Dict, Optional, Tuple, Union

    from sentry_sdk._types import Attributes, Event, Hint
    from sentry_sdk.tracing import Span


_asgi_middleware_applied = ContextVar("sentry_asgi_middleware_applied")

_DEFAULT_TRANSACTION_NAME = "generic ASGI request"

TRANSACTION_STYLE_VALUES = ("endpoint", "url")


# Vendored: https://github.com/Kludex/uvicorn/blob/b224045f5900b7f766743bcb16ba9fc3adea2606/uvicorn/_compat.py#L10-L13
if sys.version_info >= (3, 14):
    from inspect import iscoroutinefunction
else:
    from asyncio import iscoroutinefunction


def _capture_exception(exc: "Any", mechanism_type: str = "asgi") -> None:
    event, hint = event_from_exception(
        exc,
        client_options=sentry_sdk.get_client().options,
        mechanism={"type": mechanism_type, "handled": False},
    )
    sentry_sdk.capture_event(event, hint=hint)


def _looks_like_asgi3(app: "Any") -> bool:
    """
    Try to figure out if an application object supports ASGI3.

    This is how uvicorn figures out the application version as well.
    """
    if inspect.isclass(app):
        return hasattr(app, "__await__")
    elif inspect.isfunction(app):
        return iscoroutinefunction(app)
    else:
        call = getattr(app, "__call__", None)  # noqa
        return iscoroutinefunction(call)


class SentryAsgiMiddleware:
    __slots__ = (
        "app",
        "__call__",
        "transaction_style",
        "mechanism_type",
        "span_origin",
        "http_methods_to_capture",
        "root_path_in_path",
    )

    def __init__(
        self,
        app: "Any",
        unsafe_context_data: bool = False,
        transaction_style: str = "endpoint",
        mechanism_type: str = "asgi",
        span_origin: str = "manual",
        http_methods_to_capture: "Tuple[str, ...]" = DEFAULT_HTTP_METHODS_TO_CAPTURE,
        asgi_version: "Optional[int]" = None,
        root_path_in_path: "_RootPathInPath" = _RootPathInPath.EXCLUDED,
    ) -> None:
        """
        Instrument an ASGI application with Sentry. Provides HTTP/websocket
        data to sent events and basic handling for exceptions bubbling up
        through the middleware.

        :param unsafe_context_data: Disable errors when a proper contextvars installation could not be found. We do not recommend changing this from the default.
        """
        if not unsafe_context_data and not HAS_REAL_CONTEXTVARS:
            # We better have contextvars or we're going to leak state between
            # requests.
            raise RuntimeError(
                "The ASGI middleware for Sentry requires Python 3.7+ "
                "or the aiocontextvars package." + CONTEXTVARS_ERROR_MESSAGE
            )
        if transaction_style not in TRANSACTION_STYLE_VALUES:
            raise ValueError(
                "Invalid value for transaction_style: %s (must be in %s)"
                % (transaction_style, TRANSACTION_STYLE_VALUES)
            )

        asgi_middleware_while_using_starlette_or_fastapi = (
            mechanism_type == "asgi" and "starlette" in _get_installed_modules()
        )
        if asgi_middleware_while_using_starlette_or_fastapi:
            logger.warning(
                "The Sentry Python SDK can now automatically support ASGI frameworks like Starlette and FastAPI. "
                "Please remove 'SentryAsgiMiddleware' from your project. "
                "See https://docs.sentry.io/platforms/python/guides/asgi/ for more information."
            )

        self.transaction_style = transaction_style
        self.mechanism_type = mechanism_type
        self.span_origin = span_origin
        self.app = app
        self.http_methods_to_capture = http_methods_to_capture
        self.root_path_in_path = root_path_in_path

        if asgi_version is None:
            if _looks_like_asgi3(app):
                asgi_version = 3
            else:
                asgi_version = 2

        if asgi_version == 3:
            self.__call__ = self._run_asgi3
        elif asgi_version == 2:
            self.__call__ = self._run_asgi2  # type: ignore

    def _capture_lifespan_exception(self, exc: Exception) -> None:
        """Capture exceptions raise in application lifespan handlers.

        The separate function is needed to support overriding in derived integrations that use different catching mechanisms.
        """
        return _capture_exception(exc=exc, mechanism_type=self.mechanism_type)

    def _capture_request_exception(self, exc: Exception) -> None:
        """Capture exceptions raised in incoming request handlers.

        The separate function is needed to support overriding in derived integrations that use different catching mechanisms.
        """
        return _capture_exception(exc=exc, mechanism_type=self.mechanism_type)

    def _run_asgi2(self, scope: "Any") -> "Any":
        async def inner(receive: "Any", send: "Any") -> "Any":
            return await self._run_app(scope, receive, send, asgi_version=2)

        return inner

    async def _run_asgi3(self, scope: "Any", receive: "Any", send: "Any") -> "Any":
        return await self._run_app(scope, receive, send, asgi_version=3)

    async def _run_app(
        self, scope: "Any", receive: "Any", send: "Any", asgi_version: int
    ) -> "Any":
        is_recursive_asgi_middleware = _asgi_middleware_applied.get(False)
        is_lifespan = scope["type"] == "lifespan"
        if is_recursive_asgi_middleware or is_lifespan:
            try:
                if asgi_version == 2:
                    return await self.app(scope)(receive, send)
                else:
                    return await self.app(scope, receive, send)

            except Exception as exc:
                suppress_chained_exceptions = (
                    sentry_sdk.get_client()
                    .options.get("_experiments", {})
                    .get("suppress_asgi_chained_exceptions", True)
                )
                if suppress_chained_exceptions:
                    self._capture_lifespan_exception(exc)
                    raise exc from None

                exc_info = sys.exc_info()
                with capture_internal_exceptions():
                    self._capture_lifespan_exception(exc)
                reraise(*exc_info)

        client = sentry_sdk.get_client()
        span_streaming = has_span_streaming_enabled(client.options)

        _asgi_middleware_applied.set(True)
        try:
            with sentry_sdk.isolation_scope() as sentry_scope:
                with track_session(sentry_scope, session_mode="request"):
                    sentry_scope.clear_breadcrumbs()
                    sentry_scope._name = "asgi"
                    processor = partial(self.event_processor, asgi_scope=scope)
                    sentry_scope.add_event_processor(processor)

                    ty = scope["type"]
                    (
                        transaction_name,
                        transaction_source,
                    ) = self._get_transaction_name_and_source(
                        self.transaction_style,
                        scope,
                    )

                    method = scope.get("method", "").upper()

                    span_ctx: "ContextManager[Union[Span, StreamedSpan, None]]"
                    if span_streaming:
                        segment: "Optional[StreamedSpan]" = None
                        attributes: "Attributes" = {
                            "sentry.segment.name.source": getattr(
                                transaction_source, "value", transaction_source
                            ),
                            "sentry.origin": self.span_origin,
                            "network.protocol.name": ty,
                        }

                        if scope.get("client") and should_send_default_pii():
                            sentry_scope.set_attribute(
                                SPANDATA.USER_IP_ADDRESS, _get_ip(scope)
                            )

                        if ty in ("http", "websocket"):
                            if (
                                ty == "websocket"
                                or method in self.http_methods_to_capture
                            ):
                                sentry_sdk.traces.continue_trace(_get_headers(scope))

                                Scope.set_custom_sampling_context({"asgi_scope": scope})

                                attributes["sentry.op"] = f"{ty}.server"
                                segment = sentry_sdk.traces.start_span(
                                    name=transaction_name,
                                    attributes=attributes,
                                    parent_span=None,
                                )
                        else:
                            sentry_sdk.traces.new_trace()

                            Scope.set_custom_sampling_context({"asgi_scope": scope})

                            attributes["sentry.op"] = OP.HTTP_SERVER
                            segment = sentry_sdk.traces.start_span(
                                name=transaction_name,
                                attributes=attributes,
                                parent_span=None,
                            )

                        span_ctx = segment or nullcontext()

                    else:
                        transaction = None
                        if ty in ("http", "websocket"):
                            if (
                                ty == "websocket"
                                or method in self.http_methods_to_capture
                            ):
                                transaction = continue_trace(
                                    _get_headers(scope),
                                    op="{}.server".format(ty),
                                    name=transaction_name,
                                    source=transaction_source,
                                    origin=self.span_origin,
                                )
                        else:
                            transaction = Transaction(
                                op=OP.HTTP_SERVER,
                                name=transaction_name,
                                source=transaction_source,
                                origin=self.span_origin,
                            )

                        if transaction:
                            transaction.set_tag("asgi.type", ty)

                        span_ctx = (
                            sentry_sdk.start_transaction(
                                transaction,
                                custom_sampling_context={"asgi_scope": scope},
                            )
                            if transaction is not None
                            else nullcontext()
                        )

                    with span_ctx as span:
                        if isinstance(span, StreamedSpan):
                            for attribute, value in _get_request_attributes(
                                scope,
                                root_path_in_path=self.root_path_in_path,
                            ).items():
                                span.set_attribute(attribute, value)

                        try:

                            async def _sentry_wrapped_send(
                                event: "Dict[str, Any]",
                            ) -> "Any":
                                if span is not None:
                                    is_http_response = (
                                        event.get("type") == "http.response.start"
                                        and "status" in event
                                    )
                                    if is_http_response:
                                        if isinstance(span, StreamedSpan):
                                            span.status = (
                                                "error"
                                                if event["status"] >= 400
                                                else "ok"
                                            )
                                            span.set_attribute(
                                                "http.response.status_code",
                                                event["status"],
                                            )
                                        else:
                                            span.set_http_status(event["status"])

                                return await send(event)

                            if asgi_version == 2:
                                return await self.app(scope)(
                                    receive, _sentry_wrapped_send
                                )
                            else:
                                return await self.app(
                                    scope, receive, _sentry_wrapped_send
                                )

                        except Exception as exc:
                            suppress_chained_exceptions = (
                                sentry_sdk.get_client()
                                .options.get("_experiments", {})
                                .get("suppress_asgi_chained_exceptions", True)
                            )
                            if suppress_chained_exceptions:
                                self._capture_request_exception(exc)
                                raise exc from None

                            exc_info = sys.exc_info()
                            with capture_internal_exceptions():
                                self._capture_request_exception(exc)
                            reraise(*exc_info)

                        finally:
                            if isinstance(span, StreamedSpan):
                                already_set = (
                                    span is not None
                                    and span.name != _DEFAULT_TRANSACTION_NAME
                                    and span.get_attributes().get(
                                        "sentry.segment.name.source"
                                    )
                                    in [
                                        SegmentNameSource.COMPONENT.value,
                                        SegmentNameSource.ROUTE.value,
                                        SegmentNameSource.CUSTOM.value,
                                    ]
                                )
                                with capture_internal_exceptions():
                                    if not already_set:
                                        name, source = (
                                            self._get_segment_name_and_source(
                                                self.transaction_style, scope
                                            )
                                        )
                                        span.name = name
                                        span.set_attribute(
                                            "sentry.segment.name.source", source
                                        )
        finally:
            _asgi_middleware_applied.set(False)

    def event_processor(
        self, event: "Event", hint: "Hint", asgi_scope: "Any"
    ) -> "Optional[Event]":
        request_data = event.get("request", {})
        request_data.update(
            _get_request_data(asgi_scope, root_path_in_path=self.root_path_in_path)
        )
        event["request"] = deepcopy(request_data)

        # Only set transaction name if not already set by Starlette or FastAPI (or other frameworks)
        transaction = event.get("transaction")
        transaction_source = (event.get("transaction_info") or {}).get("source")
        already_set = (
            transaction is not None
            and transaction != _DEFAULT_TRANSACTION_NAME
            and transaction_source
            in [
                TransactionSource.COMPONENT,
                TransactionSource.ROUTE,
                TransactionSource.CUSTOM,
            ]
        )
        if not already_set:
            name, source = self._get_transaction_name_and_source(
                self.transaction_style, asgi_scope
            )
            event["transaction"] = name
            event["transaction_info"] = {"source": source}

        return event

    # Helper functions.
    #
    # Note: Those functions are not public API. If you want to mutate request
    # data to your liking it's recommended to use the `before_send` callback
    # for that.

    def _get_transaction_name_and_source(
        self: "SentryAsgiMiddleware", transaction_style: str, asgi_scope: "Any"
    ) -> "Tuple[str, str]":
        name = None
        source = SOURCE_FOR_STYLE[transaction_style]
        ty = asgi_scope.get("type")

        if transaction_style == "endpoint":
            endpoint = asgi_scope.get("endpoint")
            # Webframeworks like Starlette mutate the ASGI env once routing is
            # done, which is sometime after the request has started. If we have
            # an endpoint, overwrite our generic transaction name.
            if endpoint:
                name = transaction_from_function(endpoint) or ""
            else:
                name = _get_url(
                    asgi_scope,
                    "http" if ty == "http" else "ws",
                    host=None,
                    path=_get_path(
                        asgi_scope=asgi_scope, root_path_in_path=self.root_path_in_path
                    ),
                )
                source = TransactionSource.URL

        elif transaction_style == "url":
            # FastAPI includes the route object in the scope to let Sentry extract the
            # path from it for the transaction name
            route = asgi_scope.get("route")
            if route:
                path = getattr(route, "path", None)
                if path is not None:
                    name = path
            else:
                name = _get_url(
                    asgi_scope,
                    "http" if ty == "http" else "ws",
                    host=None,
                    path=_get_path(
                        asgi_scope=asgi_scope, root_path_in_path=self.root_path_in_path
                    ),
                )
                source = TransactionSource.URL

        if name is None:
            name = _DEFAULT_TRANSACTION_NAME
            source = TransactionSource.ROUTE
            return name, source

        return name, source

    def _get_segment_name_and_source(
        self: "SentryAsgiMiddleware", segment_style: str, asgi_scope: "Any"
    ) -> "Tuple[str, str]":
        name = None
        source = SEGMENT_SOURCE_FOR_STYLE[segment_style].value
        ty = asgi_scope.get("type")

        if segment_style == "endpoint":
            endpoint = asgi_scope.get("endpoint")
            # Webframeworks like Starlette mutate the ASGI env once routing is
            # done, which is sometime after the request has started. If we have
            # an endpoint, overwrite our generic transaction name.
            if endpoint:
                name = qualname_from_function(endpoint) or ""
            else:
                name = _get_url(
                    asgi_scope,
                    "http" if ty == "http" else "ws",
                    host=None,
                    path=_get_path(
                        asgi_scope=asgi_scope, root_path_in_path=self.root_path_in_path
                    ),
                )
                source = SegmentNameSource.URL.value

        elif segment_style == "url":
            # FastAPI includes the route object in the scope to let Sentry extract the
            # path from it for the transaction name
            route = asgi_scope.get("route")
            if route:
                path = getattr(route, "path", None)
                if path is not None:
                    name = path
            else:
                name = _get_url(
                    asgi_scope,
                    "http" if ty == "http" else "ws",
                    host=None,
                    path=_get_path(
                        asgi_scope=asgi_scope, root_path_in_path=self.root_path_in_path
                    ),
                )
                source = SegmentNameSource.URL.value

        if name is None:
            name = _DEFAULT_TRANSACTION_NAME
            source = SegmentNameSource.ROUTE.value
            return name, source

        return name, source



import copy
import os
from unittest import mock

import pytest

import sentry_sdk
from sentry_sdk import (
    capture_exception,
    isolation_scope,
    new_scope,
)
from sentry_sdk.client import Client, NonRecordingClient
from sentry_sdk.scope import (
    Scope,
    ScopeType,
    register_external_propagation_context,
    remove_external_propagation_context,
    should_send_default_pii,
    use_isolation_scope,
    use_scope,
)


def test_copying():
    s1 = Scope()
    s1.fingerprint = {}
    s1.set_tag("foo", "bar")

    s2 = copy.copy(s1)
    assert "foo" in s2._tags

    s1.set_tag("bam", "baz")
    assert "bam" in s1._tags
    assert "bam" not in s2._tags

    assert s1._fingerprint is s2._fingerprint


def test_all_slots_copied():
    scope = Scope()
    scope_copy = copy.copy(scope)

    # Check all attributes are copied
    for attr in set(Scope.__slots__):
        assert getattr(scope_copy, attr) == getattr(scope, attr)


def test_scope_flags_copy():
    # Assert forking creates a deepcopy of the flag buffer. The new
    # scope is free to mutate without consequence to the old scope. The
    # old scope is free to mutate without consequence to the new scope.
    old_scope = Scope()
    old_scope.flags.set("a", True)

    new_scope = old_scope.fork()
    new_scope.flags.set("a", False)
    old_scope.flags.set("b", True)
    new_scope.flags.set("c", True)

    assert old_scope.flags.get() == [
        {"flag": "a", "result": True},
        {"flag": "b", "result": True},
    ]
    assert new_scope.flags.get() == [
        {"flag": "a", "result": False},
        {"flag": "c", "result": True},
    ]


def test_set_user(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    sentry_sdk.get_isolation_scope().set_user({"id": "42", "email": "bob@example.com"})
    capture_exception(NameError())
    assert events[-1]["user"] == {"id": "42", "email": "bob@example.com"}

    sentry_sdk.get_isolation_scope().set_user(None)
    capture_exception(NameError())
    assert "user" not in events[-1]


def test_set_user_none_values_are_dropped_when_copying_to_attributes(
    sentry_init, capture_items
):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=True,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    with sentry_sdk.traces.start_span(name="test_segment"):
        sentry_sdk.get_isolation_scope().set_user(
            {"email": "ada@beans.com", "username": None}
        )
        capture_exception(NameError())

    sentry_sdk.flush()

    segment = items[0].payload

    assert "user.name" not in segment["attributes"]
    assert segment["attributes"]["user.email"] == "ada@beans.com"


def test_merging(sentry_init, capture_events):
    sentry_init()

    s = Scope()
    s.set_user({"id": "42"})

    events = capture_events()

    capture_exception(NameError(), scope=s)

    (event,) = events
    assert event["user"] == {"id": "42"}


def test_common_args():
    s = Scope()
    s.update_from_kwargs(
        user={"id": 23},
        level="warning",
        extras={"k": "v"},
        contexts={"os": {"name": "Blafasel"}},
        tags={"x": "y"},
        fingerprint=["foo"],
    )

    s2 = Scope()
    s2.set_extra("foo", "bar")
    s2.set_tag("a", "b")
    s2.set_context("device", {"a": "b"})
    s2.update_from_scope(s)

    assert s._user == {"id": 23}
    assert s._level == "warning"
    assert s._extras == {"k": "v"}
    assert s._contexts == {"os": {"name": "Blafasel"}}
    assert s._tags == {"x": "y"}
    assert s._fingerprint == ["foo"]

    assert s._user == s2._user
    assert s._level == s2._level
    assert s._fingerprint == s2._fingerprint
    assert s2._extras == {"k": "v", "foo": "bar"}
    assert s2._tags == {"a": "b", "x": "y"}
    assert s2._contexts == {"os": {"name": "Blafasel"}, "device": {"a": "b"}}


BAGGAGE_VALUE = (
    "other-vendor-value-1=foo;bar;baz, sentry-trace_id=771a43a4192642f0b136d5159a501700, "
    "sentry-public_key=49d0f7386ad645858ae85020e393bef3, sentry-sample_rate=0.01337, "
    "sentry-user_id=Am%C3%A9lie, other-vendor-value-2=foo;bar;"
)

SENTRY_TRACE_VALUE = "771a43a4192642f0b136d5159a501700-1234567890abcdef-1"


@pytest.mark.parametrize(
    "env,excepted_value",
    [
        (
            {
                "SENTRY_TRACE": SENTRY_TRACE_VALUE,
            },
            {
                "sentry-trace": SENTRY_TRACE_VALUE,
            },
        ),
        (
            {
                "SENTRY_BAGGAGE": BAGGAGE_VALUE,
            },
            {
                "baggage": BAGGAGE_VALUE,
            },
        ),
        (
            {
                "SENTRY_TRACE": SENTRY_TRACE_VALUE,
                "SENTRY_BAGGAGE": BAGGAGE_VALUE,
            },
            {
                "sentry-trace": SENTRY_TRACE_VALUE,
                "baggage": BAGGAGE_VALUE,
            },
        ),
        (
            {
                "SENTRY_USE_ENVIRONMENT": "",
                "SENTRY_TRACE": SENTRY_TRACE_VALUE,
                "SENTRY_BAGGAGE": BAGGAGE_VALUE,
            },
            {
                "sentry-trace": SENTRY_TRACE_VALUE,
                "baggage": BAGGAGE_VALUE,
            },
        ),
        (
            {
                "SENTRY_USE_ENVIRONMENT": "True",
                "SENTRY_TRACE": SENTRY_TRACE_VALUE,
                "SENTRY_BAGGAGE": BAGGAGE_VALUE,
            },
            {
                "sentry-trace": SENTRY_TRACE_VALUE,
                "baggage": BAGGAGE_VALUE,
            },
        ),
        (
            {
                "SENTRY_USE_ENVIRONMENT": "no",
                "SENTRY_TRACE": SENTRY_TRACE_VALUE,
                "SENTRY_BAGGAGE": BAGGAGE_VALUE,
            },
            None,
        ),
        (
            {
                "SENTRY_USE_ENVIRONMENT": "True",
                "MY_OTHER_VALUE": "asdf",
                "SENTRY_RELEASE": "1.0.0",
            },
            None,
        ),
    ],
)
def test_load_trace_data_from_env(env, excepted_value):
    new_env = os.environ.copy()
    new_env.update(env)

    with mock.patch.dict(os.environ, new_env):
        s = Scope()
        incoming_trace_data = s._load_trace_data_from_env()
        assert incoming_trace_data == excepted_value


def test_scope_client():
    scope = Scope(ty="test_something")
    assert scope._type == "test_something"
    assert scope.client is not None
    assert scope.client.__class__ == NonRecordingClient

    custom_client = Client()
    scope = Scope(ty="test_more", client=custom_client)
    assert scope._type == "test_more"
    assert scope.client is not None
    assert scope.client.__class__ == Client
    assert scope.client == custom_client


def test_get_current_scope():
    scope = Scope.get_current_scope()
    assert scope is not None
    assert scope.__class__ == Scope
    assert scope._type == ScopeType.CURRENT


def test_get_isolation_scope():
    scope = Scope.get_isolation_scope()
    assert scope is not None
    assert scope.__class__ == Scope
    assert scope._type == ScopeType.ISOLATION


def test_get_global_scope():
    scope = Scope.get_global_scope()
    assert scope is not None
    assert scope.__class__ == Scope
    assert scope._type == ScopeType.GLOBAL


def test_get_client():
    client = Scope.get_client()
    assert client is not None
    assert client.__class__ == NonRecordingClient
    assert not client.is_active()


def test_set_client():
    client1 = Client()
    client2 = Client()
    client3 = Client()

    current_scope = Scope.get_current_scope()
    isolation_scope = Scope.get_isolation_scope()
    global_scope = Scope.get_global_scope()

    current_scope.set_client(client1)
    isolation_scope.set_client(client2)
    global_scope.set_client(client3)

    client = Scope.get_client()
    assert client == client1

    current_scope.set_client(None)
    isolation_scope.set_client(client2)
    global_scope.set_client(client3)

    client = Scope.get_client()
    assert client == client2

    current_scope.set_client(None)
    isolation_scope.set_client(None)
    global_scope.set_client(client3)

    client = Scope.get_client()
    assert client == client3


def test_fork():
    scope = Scope()
    forked_scope = scope.fork()

    assert scope != forked_scope


def test_get_global_scope_tags():
    global_scope1 = Scope.get_global_scope()
    global_scope2 = Scope.get_global_scope()
    assert global_scope1 == global_scope2
    assert global_scope1.client.__class__ == NonRecordingClient
    assert not global_scope1.client.is_active()
    assert global_scope2.client.__class__ == NonRecordingClient
    assert not global_scope2.client.is_active()

    global_scope1.set_tag("tag1", "value")
    tags_scope1 = global_scope1._tags
    tags_scope2 = global_scope2._tags
    assert tags_scope1 == tags_scope2 == {"tag1": "value"}
    assert global_scope1.client.__class__ == NonRecordingClient
    assert not global_scope1.client.is_active()
    assert global_scope2.client.__class__ == NonRecordingClient
    assert not global_scope2.client.is_active()


def test_get_global_with_scope():
    original_global_scope = Scope.get_global_scope()

    with new_scope() as scope:
        in_with_global_scope = Scope.get_global_scope()

        assert scope is not in_with_global_scope
        assert in_with_global_scope is original_global_scope

    after_with_global_scope = Scope.get_global_scope()
    assert after_with_global_scope is original_global_scope


def test_get_global_with_isolation_scope():
    original_global_scope = Scope.get_global_scope()

    with isolation_scope() as scope:
        in_with_global_scope = Scope.get_global_scope()

        assert scope is not in_with_global_scope
        assert in_with_global_scope is original_global_scope

    after_with_global_scope = Scope.get_global_scope()
    assert after_with_global_scope is original_global_scope


def test_get_isolation_scope_tags():
    isolation_scope1 = Scope.get_isolation_scope()
    isolation_scope2 = Scope.get_isolation_scope()
    assert isolation_scope1 == isolation_scope2
    assert isolation_scope1.client.__class__ == NonRecordingClient
    assert not isolation_scope1.client.is_active()
    assert isolation_scope2.client.__class__ == NonRecordingClient
    assert not isolation_scope2.client.is_active()

    isolation_scope1.set_tag("tag1", "value")
    tags_scope1 = isolation_scope1._tags
    tags_scope2 = isolation_scope2._tags
    assert tags_scope1 == tags_scope2 == {"tag1": "value"}
    assert isolation_scope1.client.__class__ == NonRecordingClient
    assert not isolation_scope1.client.is_active()
    assert isolation_scope2.client.__class__ == NonRecordingClient
    assert not isolation_scope2.client.is_active()


def test_get_current_scope_tags():
    scope1 = Scope.get_current_scope()
    scope2 = Scope.get_current_scope()
    assert id(scope1) == id(scope2)
    assert scope1.client.__class__ == NonRecordingClient
    assert not scope1.client.is_active()
    assert scope2.client.__class__ == NonRecordingClient
    assert not scope2.client.is_active()

    scope1.set_tag("tag1", "value")
    tags_scope1 = scope1._tags
    tags_scope2 = scope2._tags
    assert tags_scope1 == tags_scope2 == {"tag1": "value"}
    assert scope1.client.__class__ == NonRecordingClient
    assert not scope1.client.is_active()
    assert scope2.client.__class__ == NonRecordingClient
    assert not scope2.client.is_active()


def test_with_isolation_scope():
    original_current_scope = Scope.get_current_scope()
    original_isolation_scope = Scope.get_isolation_scope()

    with isolation_scope() as scope:
        assert scope._type == ScopeType.ISOLATION

        in_with_current_scope = Scope.get_current_scope()
        in_with_isolation_scope = Scope.get_isolation_scope()

        assert scope is in_with_isolation_scope
        assert in_with_current_scope is not original_current_scope
        assert in_with_isolation_scope is not original_isolation_scope

    after_with_current_scope = Scope.get_current_scope()
    after_with_isolation_scope = Scope.get_isolation_scope()
    assert after_with_current_scope is original_current_scope
    assert after_with_isolation_scope is original_isolation_scope


def test_with_isolation_scope_data():
    """
    When doing `with isolation_scope()` the isolation *and* the current scope are forked,
    to prevent that by setting tags on the current scope in the context manager, data
    bleeds to the outer current scope.
    """
    isolation_scope_before = Scope.get_isolation_scope()
    current_scope_before = Scope.get_current_scope()

    isolation_scope_before.set_tag("before_isolation_scope", 1)
    current_scope_before.set_tag("before_current_scope", 1)

    with isolation_scope() as scope:
        assert scope._type == ScopeType.ISOLATION

        isolation_scope_in = Scope.get_isolation_scope()
        current_scope_in = Scope.get_current_scope()

        assert isolation_scope_in._tags == {"before_isolation_scope": 1}
        assert current_scope_in._tags == {"before_current_scope": 1}
        assert scope._tags == {"before_isolation_scope": 1}

        scope.set_tag("in_with_scope", 1)

        assert isolation_scope_in._tags == {
            "before_isolation_scope": 1,
            "in_with_scope": 1,
        }
        assert current_scope_in._tags == {"before_current_scope": 1}
        assert scope._tags == {"before_isolation_scope": 1, "in_with_scope": 1}

        isolation_scope_in.set_tag("in_with_isolation_scope", 1)

        assert isolation_scope_in._tags == {
            "before_isolation_scope": 1,
            "in_with_scope": 1,
            "in_with_isolation_scope": 1,
        }
        assert current_scope_in._tags == {"before_current_scope": 1}
        assert scope._tags == {
            "before_isolation_scope": 1,
            "in_with_scope": 1,
            "in_with_isolation_scope": 1,
        }

        current_scope_in.set_tag("in_with_current_scope", 1)

        assert isolation_scope_in._tags == {
            "before_isolation_scope": 1,
            "in_with_scope": 1,
            "in_with_isolation_scope": 1,
        }
        assert current_scope_in._tags == {
            "before_current_scope": 1,
            "in_with_current_scope": 1,
        }
        assert scope._tags == {
            "before_isolation_scope": 1,
            "in_with_scope": 1,
            "in_with_isolation_scope": 1,
        }

    isolation_scope_after = Scope.get_isolation_scope()
    current_scope_after = Scope.get_current_scope()

    isolation_scope_after.set_tag("after_isolation_scope", 1)

    assert isolation_scope_after._tags == {
        "before_isolation_scope": 1,
        "after_isolation_scope": 1,
    }
    assert current_scope_after._tags == {"before_current_scope": 1}

    current_scope_after.set_tag("after_current_scope", 1)

    assert isolation_scope_after._tags == {
        "before_isolation_scope": 1,
        "after_isolation_scope": 1,
    }
    assert current_scope_after._tags == {
        "before_current_scope": 1,
        "after_current_scope": 1,
    }


def test_with_use_isolation_scope():
    original_isolation_scope = Scope.get_isolation_scope()
    original_current_scope = Scope.get_current_scope()
    custom_isolation_scope = Scope()

    with use_isolation_scope(custom_isolation_scope) as scope:
        assert scope._type is None  # our custom scope has not type set

        in_with_isolation_scope = Scope.get_isolation_scope()
        in_with_current_scope = Scope.get_current_scope()

        assert scope is custom_isolation_scope
        assert scope is in_with_isolation_scope
        assert scope is not in_with_current_scope
        assert scope is not original_isolation_scope
        assert scope is not original_current_scope
        assert in_with_isolation_scope is not original_isolation_scope
        assert in_with_current_scope is not original_current_scope

    after_with_current_scope = Scope.get_current_scope()
    after_with_isolation_scope = Scope.get_isolation_scope()

    assert after_with_isolation_scope is original_isolation_scope
    assert after_with_current_scope is original_current_scope
    assert after_with_isolation_scope is not custom_isolation_scope
    assert after_with_current_scope is not custom_isolation_scope


def test_with_use_isolation_scope_data():
    isolation_scope_before = Scope.get_isolation_scope()
    current_scope_before = Scope.get_current_scope()
    custom_isolation_scope = Scope()

    isolation_scope_before.set_tag("before_isolation_scope", 1)
    current_scope_before.set_tag("before_current_scope", 1)
    custom_isolation_scope.set_tag("before_custom_isolation_scope", 1)

    with use_isolation_scope(custom_isolation_scope) as scope:
        assert scope._type is None  # our custom scope has not type set

        isolation_scope_in = Scope.get_isolation_scope()
        current_scope_in = Scope.get_current_scope()

        assert isolation_scope_in._tags == {"before_custom_isolation_scope": 1}
        assert current_scope_in._tags == {"before_current_scope": 1}
        assert scope._tags == {"before_custom_isolation_scope": 1}

        scope.set_tag("in_with_scope", 1)

        assert isolation_scope_in._tags == {
            "before_custom_isolation_scope": 1,
            "in_with_scope": 1,
        }
        assert current_scope_in._tags == {"before_current_scope": 1}
        assert scope._tags == {"before_custom_isolation_scope": 1, "in_with_scope": 1}

        isolation_scope_in.set_tag("in_with_isolation_scope", 1)

        assert isolation_scope_in._tags == {
            "before_custom_isolation_scope": 1,
            "in_with_scope": 1,
            "in_with_isolation_scope": 1,
        }
        assert current_scope_in._tags == {"before_current_scope": 1}
        assert scope._tags == {
            "before_custom_isolation_scope": 1,
            "in_with_scope": 1,
            "in_with_isolation_scope": 1,
        }

        current_scope_in.set_tag("in_with_current_scope", 1)

        assert isolation_scope_in._tags == {
            "before_custom_isolation_scope": 1,
            "in_with_scope": 1,
            "in_with_isolation_scope": 1,
        }
        assert current_scope_in._tags == {
            "before_current_scope": 1,
            "in_with_current_scope": 1,
        }
        assert scope._tags == {
            "before_custom_isolation_scope": 1,
            "in_with_scope": 1,
            "in_with_isolation_scope": 1,
        }

    assert custom_isolation_scope._tags == {
        "before_custom_isolation_scope": 1,
        "in_with_scope": 1,
        "in_with_isolation_scope": 1,
    }
    isolation_scope_after = Scope.get_isolation_scope()
    current_scope_after = Scope.get_current_scope()

    isolation_scope_after.set_tag("after_isolation_scope", 1)

    assert isolation_scope_after._tags == {
        "before_isolation_scope": 1,
        "after_isolation_scope": 1,
    }
    assert current_scope_after._tags == {"before_current_scope": 1}
    assert custom_isolation_scope._tags == {
        "before_custom_isolation_scope": 1,
        "in_with_scope": 1,
        "in_with_isolation_scope": 1,
    }

    current_scope_after.set_tag("after_current_scope", 1)

    assert isolation_scope_after._tags == {
        "before_isolation_scope": 1,
        "after_isolation_scope": 1,
    }
    assert current_scope_after._tags == {
        "before_current_scope": 1,
        "after_current_scope": 1,
    }
    assert custom_isolation_scope._tags == {
        "before_custom_isolation_scope": 1,
        "in_with_scope": 1,
        "in_with_isolation_scope": 1,
    }


def test_with_new_scope():
    original_current_scope = Scope.get_current_scope()
    original_isolation_scope = Scope.get_isolation_scope()

    with new_scope() as scope:
        assert scope._type == ScopeType.CURRENT

        in_with_current_scope = Scope.get_current_scope()
        in_with_isolation_scope = Scope.get_isolation_scope()

        assert scope is in_with_current_scope
        assert in_with_current_scope is not original_current_scope
        assert in_with_isolation_scope is original_isolation_scope

    after_with_current_scope = Scope.get_current_scope()
    after_with_isolation_scope = Scope.get_isolation_scope()
    assert after_with_current_scope is original_current_scope
    assert after_with_isolation_scope is original_isolation_scope


def test_with_new_scope_data():
    """
    When doing `with new_scope()` the current scope is forked but the isolation
    scope stays untouched.
    """
    isolation_scope_before = Scope.get_isolation_scope()
    current_scope_before = Scope.get_current_scope()

    isolation_scope_before.set_tag("before_isolation_scope", 1)
    current_scope_before.set_tag("before_current_scope", 1)

    with new_scope() as scope:
        assert scope._type == ScopeType.CURRENT

        isolation_scope_in = Scope.get_isolation_scope()
        current_scope_in = Scope.get_current_scope()

        assert isolation_scope_in._tags == {"before_isolation_scope": 1}
        assert current_scope_in._tags == {"before_current_scope": 1}
        assert scope._tags == {"before_current_scope": 1}

        scope.set_tag("in_with_scope", 1)

        assert isolation_scope_in._tags == {"before_isolation_scope": 1}
        assert current_scope_in._tags == {"before_current_scope": 1, "in_with_scope": 1}
        assert scope._tags == {"before_current_scope": 1, "in_with_scope": 1}

        isolation_scope_in.set_tag("in_with_isolation_scope", 1)

        assert isolation_scope_in._tags == {
            "before_isolation_scope": 1,
            "in_with_isolation_scope": 1,
        }
        assert current_scope_in._tags == {"before_current_scope": 1, "in_with_scope": 1}
        assert scope._tags == {"before_current_scope": 1, "in_with_scope": 1}

        current_scope_in.set_tag("in_with_current_scope", 1)

        assert isolation_scope_in._tags == {
            "before_isolation_scope": 1,
            "in_with_isolation_scope": 1,
        }
        assert current_scope_in._tags == {
            "before_current_scope": 1,
            "in_with_scope": 1,
            "in_with_current_scope": 1,
        }
        assert scope._tags == {
            "before_current_scope": 1,
            "in_with_scope": 1,
            "in_with_current_scope": 1,
        }

    isolation_scope_after = Scope.get_isolation_scope()
    current_scope_after = Scope.get_current_scope()

    isolation_scope_after.set_tag("after_isolation_scope", 1)

    assert isolation_scope_after._tags == {
        "before_isolation_scope": 1,
        "in_with_isolation_scope": 1,
        "after_isolation_scope": 1,
    }
    assert current_scope_after._tags == {"before_current_scope": 1}

    current_scope_after.set_tag("after_current_scope", 1)

    assert isolation_scope_after._tags == {
        "before_isolation_scope": 1,
        "in_with_isolation_scope": 1,
        "after_isolation_scope": 1,
    }
    assert current_scope_after._tags == {
        "before_current_scope": 1,
        "after_current_scope": 1,
    }


def test_with_use_scope_data():
    isolation_scope_before = Scope.get_isolation_scope()
    current_scope_before = Scope.get_current_scope()
    custom_current_scope = Scope()

    isolation_scope_before.set_tag("before_isolation_scope", 1)
    current_scope_before.set_tag("before_current_scope", 1)
    custom_current_scope.set_tag("before_custom_current_scope", 1)

    with use_scope(custom_current_scope) as scope:
        assert scope._type is None  # our custom scope has not type set

        isolation_scope_in = Scope.get_isolation_scope()
        current_scope_in = Scope.get_current_scope()

        assert isolation_scope_in._tags == {"before_isolation_scope": 1}
        assert current_scope_in._tags == {"before_custom_current_scope": 1}
        assert scope._tags == {"before_custom_current_scope": 1}

        scope.set_tag("in_with_scope", 1)

        assert isolation_scope_in._tags == {"before_isolation_scope": 1}
        assert current_scope_in._tags == {
            "before_custom_current_scope": 1,
            "in_with_scope": 1,
        }
        assert scope._tags == {"before_custom_current_scope": 1, "in_with_scope": 1}

        isolation_scope_in.set_tag("in_with_isolation_scope", 1)

        assert isolation_scope_in._tags == {
            "before_isolation_scope": 1,
            "in_with_isolation_scope": 1,
        }
        assert current_scope_in._tags == {
            "before_custom_current_scope": 1,
            "in_with_scope": 1,
        }
        assert scope._tags == {"before_custom_current_scope": 1, "in_with_scope": 1}

        current_scope_in.set_tag("in_with_current_scope", 1)

        assert isolation_scope_in._tags == {
            "before_isolation_scope": 1,
            "in_with_isolation_scope": 1,
        }
        assert current_scope_in._tags == {
            "before_custom_current_scope": 1,
            "in_with_scope": 1,
            "in_with_current_scope": 1,
        }
        assert scope._tags == {
            "before_custom_current_scope": 1,
            "in_with_scope": 1,
            "in_with_current_scope": 1,
        }

    assert custom_current_scope._tags == {
        "before_custom_current_scope": 1,
        "in_with_scope": 1,
        "in_with_current_scope": 1,
    }
    isolation_scope_after = Scope.get_isolation_scope()
    current_scope_after = Scope.get_current_scope()

    isolation_scope_after.set_tag("after_isolation_scope", 1)

    assert isolation_scope_after._tags == {
        "before_isolation_scope": 1,
        "after_isolation_scope": 1,
        "in_with_isolation_scope": 1,
    }
    assert current_scope_after._tags == {"before_current_scope": 1}
    assert custom_current_scope._tags == {
        "before_custom_current_scope": 1,
        "in_with_scope": 1,
        "in_with_current_scope": 1,
    }

    current_scope_after.set_tag("after_current_scope", 1)

    assert isolation_scope_after._tags == {
        "before_isolation_scope": 1,
        "in_with_isolation_scope": 1,
        "after_isolation_scope": 1,
    }
    assert current_scope_after._tags == {
        "before_current_scope": 1,
        "after_current_scope": 1,
    }
    assert custom_current_scope._tags == {
        "before_custom_current_scope": 1,
        "in_with_scope": 1,
        "in_with_current_scope": 1,
    }


def test_nested_scopes_with_tags(sentry_init, capture_envelopes):
    sentry_init(traces_sample_rate=1.0)
    envelopes = capture_envelopes()

    with sentry_sdk.isolation_scope() as scope1:
        scope1.set_tag("isolation_scope1", 1)

        with sentry_sdk.new_scope() as scope2:
            scope2.set_tag("current_scope2", 1)

            with sentry_sdk.start_transaction(name="trx") as trx:
                trx.set_tag("trx", 1)

                with sentry_sdk.start_span(op="span1") as span1:
                    span1.set_tag("a", 1)

                    with new_scope() as scope3:
                        scope3.set_tag("current_scope3", 1)

                        with sentry_sdk.start_span(op="span2") as span2:
                            span2.set_tag("b", 1)

    (envelope,) = envelopes
    transaction = envelope.items[0].get_transaction_event()

    assert transaction["tags"] == {"isolation_scope1": 1, "current_scope2": 1, "trx": 1}
    assert transaction["spans"][0]["tags"] == {"a": 1}
    assert transaction["spans"][1]["tags"] == {"b": 1}


def test_should_send_default_pii_true(sentry_init):
    sentry_init(send_default_pii=True)

    assert should_send_default_pii() is True


def test_should_send_default_pii_false(sentry_init):
    sentry_init(send_default_pii=False)

    assert should_send_default_pii() is False


def test_should_send_default_pii_default_false(sentry_init):
    sentry_init()

    assert should_send_default_pii() is False


def test_should_send_default_pii_false_with_dsn_and_spotlight(sentry_init):
    sentry_init(dsn="http://key@localhost/1", spotlight=True)

    assert should_send_default_pii() is False


def test_should_send_default_pii_true_without_dsn_and_spotlight(sentry_init):
    sentry_init(spotlight=True)

    assert should_send_default_pii() is True


def test_set_tags():
    scope = Scope()
    scope.set_tags({"tag1": "value1", "tag2": "value2"})
    event = scope.apply_to_event({}, {})

    assert event["tags"] == {"tag1": "value1", "tag2": "value2"}, "Setting tags failed"

    scope.set_tags({"tag2": "updated", "tag3": "new"})
    event = scope.apply_to_event({}, {})

    assert event["tags"] == {
        "tag1": "value1",
        "tag2": "updated",
        "tag3": "new",
    }, "Updating tags failed"

    scope.set_tags({})
    event = scope.apply_to_event({}, {})

    assert event["tags"] == {
        "tag1": "value1",
        "tag2": "updated",
        "tag3": "new",
    }, "Updating tags with empty dict changed tags"


def test_last_event_id(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    assert Scope.last_event_id() is None

    sentry_sdk.capture_exception(Exception("test"))

    assert Scope.last_event_id() is not None


def test_last_event_id_transaction(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    assert Scope.last_event_id() is None

    with sentry_sdk.start_transaction(name="test"):
        pass

    assert Scope.last_event_id() is None, "Transaction should not set last_event_id"


def test_last_event_id_cleared(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    # Make sure last_event_id is set
    sentry_sdk.capture_exception(Exception("test"))
    assert Scope.last_event_id() is not None

    # Clearing the isolation scope should clear the last_event_id
    Scope.get_isolation_scope().clear()

    assert Scope.last_event_id() is None, "last_event_id should be cleared"


@pytest.mark.tests_internal_exceptions
@pytest.mark.parametrize("error_cls", [LookupError, ValueError])
@pytest.mark.parametrize(
    "scope_manager",
    [
        new_scope,
        use_scope,
    ],
)
def test_handle_error_on_token_reset_current_scope(error_cls, scope_manager):
    with mock.patch("sentry_sdk.scope.capture_internal_exception") as mock_capture:
        with mock.patch("sentry_sdk.scope._current_scope") as mock_token_var:
            mock_token_var.reset.side_effect = error_cls()

            mock_token = mock.Mock()
            mock_token_var.set.return_value = mock_token

            try:
                if scope_manager == use_scope:
                    with scope_manager(Scope()):
                        pass
                else:
                    with scope_manager():
                        pass

            except Exception:
                pytest.fail(f"Context manager should handle {error_cls} gracefully")

            mock_capture.assert_called_once()
            mock_token_var.reset.assert_called_once_with(mock_token)


@pytest.mark.tests_internal_exceptions
@pytest.mark.parametrize("error_cls", [LookupError, ValueError])
@pytest.mark.parametrize(
    "scope_manager",
    [
        isolation_scope,
        use_isolation_scope,
    ],
)
def test_handle_error_on_token_reset_isolation_scope(error_cls, scope_manager):
    with mock.patch("sentry_sdk.scope.capture_internal_exception") as mock_capture:
        with mock.patch("sentry_sdk.scope._current_scope") as mock_current_scope:
            with mock.patch(
                "sentry_sdk.scope._isolation_scope"
            ) as mock_isolation_scope:
                mock_isolation_scope.reset.side_effect = error_cls()
                mock_current_token = mock.Mock()
                mock_current_scope.set.return_value = mock_current_token

                try:
                    if scope_manager == use_isolation_scope:
                        with scope_manager(Scope()):
                            pass
                    else:
                        with scope_manager():
                            pass

                except Exception:
                    pytest.fail(f"Context manager should handle {error_cls} gracefully")

                mock_capture.assert_called_once()
                mock_current_scope.reset.assert_called_once_with(mock_current_token)


def test_trace_context_tracing(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    with sentry_sdk.start_transaction(name="trx") as transaction:
        with sentry_sdk.start_span(op="span1"):
            with sentry_sdk.start_span(op="span2") as span:
                trace_context = sentry_sdk.get_current_scope().get_trace_context()

    assert trace_context["trace_id"] == transaction.trace_id
    assert trace_context["span_id"] == span.span_id
    assert trace_context["parent_span_id"] == span.parent_span_id
    assert "dynamic_sampling_context" in trace_context


def test_trace_context_external_tracing(sentry_init):
    sentry_init()

    def external_propagation_context():
        return ("trace_id_foo", "span_id_bar")

    register_external_propagation_context(external_propagation_context)

    scope = sentry_sdk.get_current_scope()

    trace_context = scope.get_trace_context()
    assert trace_context["trace_id"] == "trace_id_foo"
    assert trace_context["span_id"] == "span_id_bar"

    headers = list(scope.iter_trace_propagation_headers())
    assert not headers

    remove_external_propagation_context()


def test_trace_context_without_performance(sentry_init):
    sentry_init()

    with sentry_sdk.isolation_scope() as isolation_scope:
        trace_context = sentry_sdk.get_current_scope().get_trace_context()

    propagation_context = isolation_scope._propagation_context
    assert propagation_context is not None
    assert trace_context["trace_id"] == propagation_context.trace_id
    assert trace_context["span_id"] == propagation_context.span_id
    assert trace_context["parent_span_id"] == propagation_context.parent_span_id
    assert "dynamic_sampling_context" in trace_context


def test_conversation_id_set_get():
    """Test that set_conversation_id and get_conversation_id work correctly."""
    scope = Scope()
    assert scope.get_conversation_id() is None

    scope.set_conversation_id("test-conv-123")
    assert scope.get_conversation_id() == "test-conv-123"


def test_conversation_id_remove():
    """Test that remove_conversation_id clears the conversation ID."""
    scope = Scope()
    scope.set_conversation_id("test-conv-456")
    assert scope.get_conversation_id() == "test-conv-456"

    scope.remove_conversation_id()
    assert scope.get_conversation_id() is None


def test_conversation_id_overwrite():
    """Test that set_conversation_id overwrites existing value."""
    scope = Scope()
    scope.set_conversation_id("first-conv")
    scope.set_conversation_id("second-conv")
    assert scope.get_conversation_id() == "second-conv"


def test_conversation_id_copy():
    """Test that conversation_id is preserved when scope is copied."""
    scope1 = Scope()
    scope1.set_conversation_id("copy-test-conv")

    scope2 = copy.copy(scope1)
    assert scope2.get_conversation_id() == "copy-test-conv"

    # Modifying copy should not affect original
    scope2.set_conversation_id("modified-conv")
    assert scope1.get_conversation_id() == "copy-test-conv"
    assert scope2.get_conversation_id() == "modified-conv"


def test_conversation_id_clear():
    """Test that conversation_id is cleared when scope.clear() is called."""
    scope = Scope()
    scope.set_conversation_id("clear-test-conv")
    assert scope.get_conversation_id() == "clear-test-conv"

    scope.clear()
    assert scope.get_conversation_id() is None



import warnings

import pytest

import sentry_sdk
from sentry_sdk.data_collection import _ALL_HTTP_BODY_TYPES
from sentry_sdk.utils import has_data_collection_enabled


def test_kvcb_invalid_mode():
    with pytest.raises(ValueError):
        sentry_sdk.init(_experiments={"data_collection": {"cookies": {"mode": "nope"}}})  # type: ignore Purposely ignoring to test invalid option


def test_kvcb_from_dict_defaults_mode():
    sentry_sdk.init(
        _experiments={
            "data_collection": {"cookies": {"mode": "denylist", "terms": ["x"]}}
        }
    )
    client = sentry_sdk.get_client()
    assert client.options["data_collection"]["cookies"] == {
        "mode": "denylist",
        "terms": ["x"],
    }


def test_http_headers_collection_defaults():
    default_terms = ["forwarded", "-ip", "remote-", "via", "-user"]

    sentry_sdk.init(_experiments={"data_collection": {"http_headers": {}}})  # type: ignore Purposely ignoring to test invalid option
    client = sentry_sdk.get_client()
    assert client.options["data_collection"]["http_headers"]["request"] == {
        "mode": "denylist"
    }

    sentry_sdk.init(_experiments={"data_collection": {"http_headers": "off"}})  # type: ignore Purposely ignoring to test invalid option
    client = sentry_sdk.get_client()
    assert client.options["data_collection"]["http_headers"]["request"] == {
        "mode": "denylist"
    }

    sentry_sdk.init()
    client = sentry_sdk.get_client()
    assert client.options["data_collection"]["http_headers"]["request"] == {
        "mode": "denylist",
        "terms": default_terms,
    }


def test_http_headers_use_default_in_setting_with_missing_config():
    sentry_sdk.init(
        _experiments={
            "data_collection": {
                "http_headers": {
                    "request": {"mode": "allowlist", "terms": ["x-id"]},
                }
            }
        }
    )

    client = sentry_sdk.get_client()

    assert client.options["data_collection"]["http_headers"]["request"] == {
        "mode": "allowlist",
        "terms": ["x-id"],
    }


def _initialize_client_with_config(**options):
    sentry_sdk.init(**options)
    return sentry_sdk.get_client().options["data_collection"]


def _get(dc, path):
    obj = dc
    for part in path.split("."):
        obj = obj[part]
    return obj


@pytest.mark.parametrize(
    "options,expected",
    [
        pytest.param(
            {},
            {
                "user_info": False,
                "gen_ai.inputs": False,
                "gen_ai.outputs": False,
                "cookies.mode": "off",
                "query_params.mode": "off",
                "http_headers.request.mode": "denylist",
                "http_bodies": _ALL_HTTP_BODY_TYPES,
                "queues": False,
                "frame_context_lines": 5,
            },
            id="no_options_collects_no_pii",
        ),
        pytest.param(
            {"send_default_pii": True},
            {
                "user_info": True,
                "gen_ai.inputs": True,
                "gen_ai.outputs": True,
                "cookies.mode": "denylist",
                "query_params.mode": "denylist",
                "queues": True,
            },
            id="send_default_pii_true_collects_pii",
        ),
        pytest.param(
            {"send_default_pii": False},
            {"user_info": False, "cookies.mode": "off", "queues": False},
            id="send_default_pii_false_collects_no_pii",
        ),
        pytest.param(
            {"_experiments": {"data_collection": {}}},
            {
                "user_info": True,
                "gen_ai.inputs": True,
                "gen_ai.outputs": True,
                "cookies.mode": "denylist",
                "query_params.mode": "denylist",
                "http_bodies": _ALL_HTTP_BODY_TYPES,
                "queues": True,
            },
            id="explicit_data_collection_uses_spec_defaults",
        ),
        pytest.param(
            {
                "_experiments": {
                    "data_collection": {"user_info": False, "http_bodies": []}
                }
            },
            {
                "user_info": False,
                "http_bodies": [],
                "gen_ai.inputs": True,
                "cookies.mode": "denylist",
            },
            id="explicit_partial_fills_omitted_with_spec_defaults",
        ),
        pytest.param(
            {
                "_experiments": {"data_collection": {}},
                "include_local_variables": False,
                "include_source_context": False,
            },
            {"stack_frame_variables": False, "frame_context_lines": 0},
            id="explicit_frame_fields_fall_back_to_legacy_options",
        ),
        pytest.param(
            {
                "_experiments": {
                    "data_collection": {
                        "cookies": {"mode": "off"},
                        "query_params": {"mode": "allowlist", "terms": ["page"]},
                        "http_headers": {"request": {"mode": "off"}},
                        "gen_ai": {"inputs": False, "outputs": True},
                    }
                }
            },
            {
                "cookies.mode": "off",
                "query_params.mode": "allowlist",
                "query_params.terms": ["page"],
                "http_headers.request.mode": "off",
                "gen_ai.inputs": False,
                "gen_ai.outputs": True,
            },
            id="accepts_dict_with_nested_dicts",
        ),
        pytest.param(
            {
                "_experiments": {
                    "data_collection": {
                        "cookies": None,
                        "http_headers": None,
                        "query_params": None,
                        "graphql": None,
                        "gen_ai": None,
                    }
                }
            },
            {
                "cookies.mode": "denylist",
                "http_headers.request.mode": "denylist",
                "query_params.mode": "denylist",
                "graphql.document": True,
                "graphql.variables": True,
                "gen_ai.inputs": True,
                "gen_ai.outputs": True,
            },
            id="none_values_fall_back_to_spec_defaults",
        ),
        pytest.param(
            {"_experiments": {"data_collection": {}}},
            {
                "graphql.document": True,
                "graphql.variables": True,
                "database_query_data": True,
            },
            id="explicit_graphql_database_defaults",
        ),
        pytest.param(
            {"send_default_pii": False},
            {
                "graphql.document": False,
                "graphql.variables": False,
                "database_query_data": False,
                "queues": False,
            },
            id="legacy_pii_off_gates_graphql_database_and_queues",
        ),
        pytest.param(
            {"send_default_pii": True},
            {
                "graphql.document": True,
                "graphql.variables": True,
                "database_query_data": True,
                "queues": True,
            },
            id="legacy_pii_on_collects_graphql_database_and_queues",
        ),
        pytest.param(
            {"_experiments": {"data_collection": {"graphql": {"variables": False}}}},
            {"graphql.document": True, "graphql.variables": False},
            id="explicit_partial_graphql_fills_omitted",
        ),
        pytest.param(
            {"_experiments": {"data_collection": {"frame_context_lines": True}}},
            {"frame_context_lines": 5},
            id="frame_context_lines_bool_fallback_true",
        ),
        pytest.param(
            {"_experiments": {"data_collection": {"frame_context_lines": False}}},
            {"frame_context_lines": 0},
            id="frame_context_lines_bool_fallback_false",
        ),
        pytest.param(
            {"_experiments": {"data_collection": {"frame_context_lines": 3}}},
            {"frame_context_lines": 3},
            id="frame_context_lines_bool_fallback_3",
        ),
        pytest.param(
            {"_experiments": {"data_collection": {"frame_context_lines": 0}}},
            {"frame_context_lines": 0},
            id="frame_context_lines_bool_fallback_0",
        ),
    ],
)
def test_initalize_client_data_collection(options, expected):
    dc = _initialize_client_with_config(**options)
    for path, value in expected.items():
        assert _get(dc, path) == value, f"{path} != {value!r}"


def test_initialize_client_data_collection_overrides_send_default_pii_and_warns():
    with warnings.catch_warnings(record=True) as caught:
        warnings.simplefilter("always")
        dc = _initialize_client_with_config(
            send_default_pii=True,
            _experiments={"data_collection": {"user_info": False}},
        )
    assert dc["user_info"] is False  # data_collection wins
    assert any(issubclass(w.category, DeprecationWarning) for w in caught)


@pytest.mark.parametrize(
    "init_kwargs,expected",
    [
        pytest.param(
            {},
            {
                "user_info": False,
                "provided_by_user": False,
                "should_send_default_pii": False,
            },
            id="defaults_to_no_pii",
        ),
        pytest.param(
            {"send_default_pii": True},
            {"user_info": True, "provided_by_user": False},
            id="send_default_pii_enables_user_info",
        ),
        pytest.param(
            {"_experiments": {"data_collection": {"user_info": False}}},
            {"user_info": False, "provided_by_user": True},
            id="explicit_data_collection_overrides_user_info",
        ),
        pytest.param(
            {"spotlight": True},
            {"provided_by_user": False, "user_info": True},
            id="dsnless_spotlight_rederives_data_collection",
        ),
        pytest.param(
            {
                "spotlight": True,
                "_experiments": {"data_collection": {"user_info": False}},
            },
            {"provided_by_user": True, "user_info": False},
            id="dsnless_spotlight_respects_explicit_data_collection",
        ),
    ],
)
def test_client_data_collection_settings(init_kwargs, expected):
    sentry_sdk.init(**init_kwargs)
    client = sentry_sdk.get_client()
    for key, value in expected.items():
        if key == "should_send_default_pii":
            assert client.should_send_default_pii() is value
        else:
            assert client.options["data_collection"][key] is value


def test_has_data_collection_enabled_gates_on_presence():
    assert has_data_collection_enabled(None) is False
    assert has_data_collection_enabled({"_experiments": {}}) is False
    assert (
        has_data_collection_enabled({"_experiments": {"data_collection": {}}}) is True
    )
    assert (
        has_data_collection_enabled({"_experiments": {"data_collection": None}}) is True
    )


def test_no_experiments_data_collection_values_fall_back_to_send_default_pii_configuration():
    sentry_sdk.init(send_default_pii=True)
    client = sentry_sdk.get_client()
    dc = client.options["data_collection"]
    assert dc["provided_by_user"] is False
    assert dc["user_info"] is True
    assert has_data_collection_enabled(client.options) is False



import os
import time

import sentry_sdk
from sentry_sdk.integrations.aws_lambda import AwsLambdaIntegration

sentry_sdk.init(
    dsn=os.environ.get("SENTRY_DSN"),
    traces_sample_rate=1.0,
    integrations=[AwsLambdaIntegration(timeout_warning=True)],
)


def handler(event, context):
    sentry_sdk.set_tag("custom_tag", "custom_value")
    time.sleep(15)
    return {
        "event": event,
    }



import asyncio
import os
from unittest import mock

import httpx2
import pytest

import sentry_sdk
from sentry_sdk import capture_message, start_transaction
from sentry_sdk.consts import MATCH_ALL, OP, SPANDATA
from sentry_sdk.integrations.httpx2 import Httpx2Integration
from tests.conftest import ApproxDict


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_crumb_capture_and_hint(
    sentry_init, capture_events, httpx2_client, httpx2_mock
):
    httpx2_mock.add_response()

    def before_breadcrumb(crumb, hint):
        crumb["data"]["extra"] = "foo"
        return crumb

    sentry_init(integrations=[Httpx2Integration()], before_breadcrumb=before_breadcrumb)

    url = "http://example.com/"

    with start_transaction():
        events = capture_events()

        if asyncio.iscoroutinefunction(httpx2_client.get):
            response = asyncio.run(httpx2_client.get(url))
        else:
            response = httpx2_client.get(url)

        assert response.status_code == 200
        capture_message("Testing!")

        (event,) = events

        crumb = event["breadcrumbs"]["values"][0]
        assert crumb["type"] == "http"
        assert crumb["category"] == "httplib"
        assert crumb["data"] == ApproxDict(
            {
                "url": url,
                SPANDATA.HTTP_METHOD: "GET",
                SPANDATA.HTTP_FRAGMENT: "",
                SPANDATA.HTTP_QUERY: "",
                SPANDATA.HTTP_STATUS_CODE: 200,
                "reason": "OK",
                "extra": "foo",
            }
        )


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
@pytest.mark.parametrize(
    "status_code,level",
    [
        (200, None),
        (301, None),
        (403, "warning"),
        (405, "warning"),
        (500, "error"),
    ],
)
def test_crumb_capture_client_error(
    sentry_init, capture_events, httpx2_client, httpx2_mock, status_code, level
):
    httpx2_mock.add_response(status_code=status_code)

    sentry_init(integrations=[Httpx2Integration()])

    url = "http://example.com/"

    with start_transaction():
        events = capture_events()

        if asyncio.iscoroutinefunction(httpx2_client.get):
            response = asyncio.run(httpx2_client.get(url))
        else:
            response = httpx2_client.get(url)

        assert response.status_code == status_code
        capture_message("Testing!")

        (event,) = events

        crumb = event["breadcrumbs"]["values"][0]
        assert crumb["type"] == "http"
        assert crumb["category"] == "httplib"

        if level is None:
            assert "level" not in crumb
        else:
            assert crumb["level"] == level

        assert crumb["data"] == ApproxDict(
            {
                "url": url,
                SPANDATA.HTTP_METHOD: "GET",
                SPANDATA.HTTP_FRAGMENT: "",
                SPANDATA.HTTP_QUERY: "",
                SPANDATA.HTTP_STATUS_CODE: status_code,
            }
        )


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_outgoing_trace_headers_legacy(sentry_init, httpx2_client, httpx2_mock):
    httpx2_mock.add_response()

    sentry_init(
        traces_sample_rate=1.0,
        integrations=[Httpx2Integration()],
    )

    url = "http://example.com/"

    with start_transaction(
        name="/interactions/other-dogs/new-dog",
        op="greeting.sniff",
        trace_id="01234567890123456789012345678901",
    ) as transaction:
        if asyncio.iscoroutinefunction(httpx2_client.get):
            response = asyncio.run(httpx2_client.get(url))
        else:
            response = httpx2_client.get(url)

        request_span = transaction._span_recorder.spans[-1]
        assert response.request.headers[
            "sentry-trace"
        ] == "{trace_id}-{parent_span_id}-{sampled}".format(
            trace_id=transaction.trace_id,
            parent_span_id=request_span.span_id,
            sampled=1,
        )


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_outgoing_trace_headers_append_to_baggage_legacy(
    sentry_init,
    httpx2_client,
    httpx2_mock,
):
    httpx2_mock.add_response()

    sentry_init(
        traces_sample_rate=1.0,
        integrations=[Httpx2Integration()],
        release="d08ebdb9309e1b004c6f52202de58a09c2268e42",
    )

    url = "http://example.com/"

    # patch random.randrange to return a predictable sample_rand value
    with mock.patch("sentry_sdk.tracing_utils.Random.randrange", return_value=500000):
        with start_transaction(
            name="/interactions/other-dogs/new-dog",
            op="greeting.sniff",
            trace_id="01234567890123456789012345678901",
        ) as transaction:
            if asyncio.iscoroutinefunction(httpx2_client.get):
                response = asyncio.run(
                    httpx2_client.get(url, headers={"baGGage": "custom=data"})
                )
            else:
                response = httpx2_client.get(url, headers={"baGGage": "custom=data"})

            request_span = transaction._span_recorder.spans[-1]
            assert response.request.headers[
                "sentry-trace"
            ] == "{trace_id}-{parent_span_id}-{sampled}".format(
                trace_id=transaction.trace_id,
                parent_span_id=request_span.span_id,
                sampled=1,
            )
            assert (
                response.request.headers["baggage"]
                == "custom=data,sentry-trace_id=01234567890123456789012345678901,sentry-sample_rand=0.500000,sentry-environment=production,sentry-release=d08ebdb9309e1b004c6f52202de58a09c2268e42,sentry-transaction=/interactions/other-dogs/new-dog,sentry-sample_rate=1.0,sentry-sampled=true"
            )


@pytest.mark.parametrize(
    "httpx2_client,trace_propagation_targets,url,trace_propagated",
    [
        [
            httpx2.Client(),
            None,
            "https://example.com/",
            False,
        ],
        [
            httpx2.Client(),
            [],
            "https://example.com/",
            False,
        ],
        [
            httpx2.Client(),
            [MATCH_ALL],
            "https://example.com/",
            True,
        ],
        [
            httpx2.Client(),
            ["https://example.com/"],
            "https://example.com/",
            True,
        ],
        [
            httpx2.Client(),
            ["https://example.com/"],
            "https://example.com",
            False,
        ],
        [
            httpx2.Client(),
            ["https://example.com"],
            "https://example.com",
            True,
        ],
        [
            httpx2.Client(),
            ["https://example.com", r"https?:\/\/[\w\-]+(\.[\w\-]+)+\.net"],
            "https://example.net",
            False,
        ],
        [
            httpx2.Client(),
            ["https://example.com", r"https?:\/\/[\w\-]+(\.[\w\-]+)+\.net"],
            "https://good.example.net",
            True,
        ],
        [
            httpx2.Client(),
            ["https://example.com", r"https?:\/\/[\w\-]+(\.[\w\-]+)+\.net"],
            "https://good.example.net/some/thing",
            True,
        ],
        [
            httpx2.AsyncClient(),
            None,
            "https://example.com/",
            False,
        ],
        [
            httpx2.AsyncClient(),
            [],
            "https://example.com/",
            False,
        ],
        [
            httpx2.AsyncClient(),
            [MATCH_ALL],
            "https://example.com/",
            True,
        ],
        [
            httpx2.AsyncClient(),
            ["https://example.com/"],
            "https://example.com/",
            True,
        ],
        [
            httpx2.AsyncClient(),
            ["https://example.com/"],
            "https://example.com",
            False,
        ],
        [
            httpx2.AsyncClient(),
            ["https://example.com"],
            "https://example.com",
            True,
        ],
        [
            httpx2.AsyncClient(),
            ["https://example.com", r"https?:\/\/[\w\-]+(\.[\w\-]+)+\.net"],
            "https://example.net",
            False,
        ],
        [
            httpx2.AsyncClient(),
            ["https://example.com", r"https?:\/\/[\w\-]+(\.[\w\-]+)+\.net"],
            "https://good.example.net",
            True,
        ],
        [
            httpx2.AsyncClient(),
            ["https://example.com", r"https?:\/\/[\w\-]+(\.[\w\-]+)+\.net"],
            "https://good.example.net/some/thing",
            True,
        ],
    ],
)
def test_option_trace_propagation_targets(
    sentry_init,
    httpx2_client,
    httpx2_mock,  # this comes from pytest-httpx2
    trace_propagation_targets,
    url,
    trace_propagated,
):
    httpx2_mock.add_response()

    sentry_init(
        release="test",
        trace_propagation_targets=trace_propagation_targets,
        traces_sample_rate=1.0,
        integrations=[Httpx2Integration()],
    )

    # Must be in a transaction to propagate headers
    with sentry_sdk.start_transaction():
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    request_headers = httpx2_mock.get_request().headers

    if trace_propagated:
        assert "sentry-trace" in request_headers
    else:
        assert "sentry-trace" not in request_headers


def test_do_not_propagate_outside_transaction(sentry_init, httpx2_mock):
    httpx2_mock.add_response()

    sentry_init(
        traces_sample_rate=1.0,
        trace_propagation_targets=[MATCH_ALL],
        integrations=[Httpx2Integration()],
    )

    httpx2_client = httpx2.Client()
    httpx2_client.get("http://example.com/")

    request_headers = httpx2_mock.get_request().headers
    assert "sentry-trace" not in request_headers


@pytest.mark.tests_internal_exceptions
def test_omit_url_data_if_parsing_fails(sentry_init, capture_events, httpx2_mock):
    httpx2_mock.add_response()

    sentry_init(integrations=[Httpx2Integration()])

    httpx2_client = httpx2.Client()
    url = "http://example.com"

    events = capture_events()
    with mock.patch(
        "sentry_sdk.integrations.httpx2.parse_url",
        side_effect=ValueError,
    ):
        response = httpx2_client.get(url)

    assert response.status_code == 200
    capture_message("Testing!")

    (event,) = events
    assert event["breadcrumbs"]["values"][0]["data"] == ApproxDict(
        {
            SPANDATA.HTTP_METHOD: "GET",
            SPANDATA.HTTP_STATUS_CODE: 200,
            "reason": "OK",
            # no url related data
        }
    )

    assert "url" not in event["breadcrumbs"]["values"][0]["data"]
    assert SPANDATA.HTTP_FRAGMENT not in event["breadcrumbs"]["values"][0]["data"]
    assert SPANDATA.HTTP_QUERY not in event["breadcrumbs"]["values"][0]["data"]


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_request_source_disabled_legacy(
    sentry_init, capture_events, httpx2_client, httpx2_mock
):
    httpx2_mock.add_response()
    sentry_options = {
        "integrations": [Httpx2Integration()],
        "traces_sample_rate": 1.0,
        "enable_http_request_source": False,
        "http_request_source_threshold_ms": 0,
    }

    sentry_init(**sentry_options)

    events = capture_events()

    url = "http://example.com/"

    with start_transaction(name="test_transaction"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    (event,) = events

    span = event["spans"][-1]
    assert span["description"].startswith("GET")

    data = span.get("data", {})

    assert SPANDATA.CODE_LINENO not in data
    assert SPANDATA.CODE_NAMESPACE not in data
    assert SPANDATA.CODE_FILEPATH not in data
    assert SPANDATA.CODE_FUNCTION not in data


@pytest.mark.parametrize("enable_http_request_source", [None, True])
@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_request_source_enabled_legacy(
    sentry_init,
    capture_events,
    enable_http_request_source,
    httpx2_client,
    httpx2_mock,
):
    httpx2_mock.add_response()
    sentry_options = {
        "integrations": [Httpx2Integration()],
        "traces_sample_rate": 1.0,
        "http_request_source_threshold_ms": 0,
    }
    if enable_http_request_source is not None:
        sentry_options["enable_http_request_source"] = enable_http_request_source

    sentry_init(**sentry_options)

    events = capture_events()

    url = "http://example.com/"

    with start_transaction(name="test_transaction"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    (event,) = events

    span = event["spans"][-1]
    assert span["description"].startswith("GET")

    data = span.get("data", {})

    assert SPANDATA.CODE_LINENO in data
    assert SPANDATA.CODE_NAMESPACE in data
    assert SPANDATA.CODE_FILEPATH in data
    assert SPANDATA.CODE_FUNCTION in data


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_request_source_legacy(sentry_init, capture_events, httpx2_client, httpx2_mock):
    httpx2_mock.add_response()

    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=0,
    )

    events = capture_events()

    url = "http://example.com/"

    with start_transaction(name="test_transaction"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    (event,) = events

    span = event["spans"][-1]
    assert span["description"].startswith("GET")

    data = span.get("data", {})

    assert SPANDATA.CODE_LINENO in data
    assert SPANDATA.CODE_NAMESPACE in data
    assert SPANDATA.CODE_FILEPATH in data
    assert SPANDATA.CODE_FUNCTION in data

    assert type(data.get(SPANDATA.CODE_LINENO)) == int
    assert data.get(SPANDATA.CODE_LINENO) > 0
    assert data.get(SPANDATA.CODE_NAMESPACE) == "tests.integrations.httpx2.test_httpx2"
    assert data.get(SPANDATA.CODE_FILEPATH).endswith(
        "tests/integrations/httpx2/test_httpx2.py"
    )

    is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
    assert is_relative_path

    assert data.get(SPANDATA.CODE_FUNCTION) == "test_request_source_legacy"


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_request_source_with_module_in_search_path_legacy(
    sentry_init, capture_events, httpx2_client, httpx2_mock
):
    """
    Test that request source is relative to the path of the module it ran in
    """
    httpx2_mock.add_response()
    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=0,
    )

    events = capture_events()

    url = "http://example.com/"

    with start_transaction(name="test_transaction"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            from httpx2_helpers.helpers import async_get_request_with_client

            asyncio.run(async_get_request_with_client(httpx2_client, url))
        else:
            from httpx2_helpers.helpers import get_request_with_client

            get_request_with_client(httpx2_client, url)

    (event,) = events

    span = event["spans"][-1]
    assert span["description"].startswith("GET")

    data = span.get("data", {})

    assert SPANDATA.CODE_LINENO in data
    assert SPANDATA.CODE_NAMESPACE in data
    assert SPANDATA.CODE_FILEPATH in data
    assert SPANDATA.CODE_FUNCTION in data

    assert type(data.get(SPANDATA.CODE_LINENO)) == int
    assert data.get(SPANDATA.CODE_LINENO) > 0
    assert data.get(SPANDATA.CODE_NAMESPACE) == "httpx2_helpers.helpers"
    assert data.get(SPANDATA.CODE_FILEPATH) == "httpx2_helpers/helpers.py"

    is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
    assert is_relative_path

    if asyncio.iscoroutinefunction(httpx2_client.get):
        assert data.get(SPANDATA.CODE_FUNCTION) == "async_get_request_with_client"
    else:
        assert data.get(SPANDATA.CODE_FUNCTION) == "get_request_with_client"


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_no_request_source_if_duration_too_short_legacy(
    sentry_init, capture_events, httpx2_client, httpx2_mock
):
    httpx2_mock.add_response()

    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        # Threshold so high no real request will ever exceed it
        http_request_source_threshold_ms=9999999,
    )

    events = capture_events()

    url = "http://example.com/"

    with start_transaction(name="test_transaction"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    (event,) = events

    span = event["spans"][-1]
    assert span["description"].startswith("GET")

    data = span.get("data", {})

    assert SPANDATA.CODE_LINENO not in data
    assert SPANDATA.CODE_NAMESPACE not in data
    assert SPANDATA.CODE_FILEPATH not in data
    assert SPANDATA.CODE_FUNCTION not in data


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_request_source_if_duration_over_threshold_legacy(
    sentry_init, capture_events, httpx2_client, httpx2_mock
):
    httpx2_mock.add_response()

    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        # Threshold is low so any request will exceed it
        http_request_source_threshold_ms=0,
    )

    events = capture_events()

    url = "http://example.com/"

    with start_transaction(name="test_transaction"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    (event,) = events

    span = event["spans"][-1]
    assert span["description"].startswith("GET")

    data = span.get("data", {})

    assert SPANDATA.CODE_LINENO in data
    assert SPANDATA.CODE_NAMESPACE in data
    assert SPANDATA.CODE_FILEPATH in data
    assert SPANDATA.CODE_FUNCTION in data

    assert type(data.get(SPANDATA.CODE_LINENO)) == int
    assert data.get(SPANDATA.CODE_LINENO) > 0
    assert data.get(SPANDATA.CODE_NAMESPACE) == "tests.integrations.httpx2.test_httpx2"
    assert data.get(SPANDATA.CODE_FILEPATH).endswith(
        "tests/integrations/httpx2/test_httpx2.py"
    )

    is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
    assert is_relative_path

    assert (
        data.get(SPANDATA.CODE_FUNCTION)
        == "test_request_source_if_duration_over_threshold_legacy"
    )


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_span_origin_legacy(sentry_init, capture_events, httpx2_client, httpx2_mock):
    httpx2_mock.add_response()

    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
    )

    events = capture_events()

    url = "http://example.com/"

    with start_transaction(name="test_transaction"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    (event,) = events

    assert event["contexts"]["trace"]["origin"] == "manual"
    assert event["spans"][0]["origin"] == "auto.http.httpx2"


def _get_http_client_span(items):
    return next(
        item.payload
        for item in items
        if item.payload.get("attributes", {}).get("sentry.op") == OP.HTTP_CLIENT
    )


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_outgoing_trace_headers_span_streaming(
    sentry_init, capture_items, httpx2_client, httpx2_mock
):
    httpx2_mock.add_response()

    sentry_init(
        traces_sample_rate=1.0,
        integrations=[Httpx2Integration()],
        trace_lifecycle="stream",
    )

    url = "http://example.com/"

    items = capture_items("span")

    with sentry_sdk.traces.start_span(name="test"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            response = asyncio.run(httpx2_client.get(url))
        else:
            response = httpx2_client.get(url)

    sentry_sdk.flush()

    http_span = _get_http_client_span(items)

    assert response.request.headers[
        "sentry-trace"
    ] == "{trace_id}-{span_id}-{sampled}".format(
        trace_id=http_span["trace_id"],
        span_id=http_span["span_id"],
        sampled=1,
    )


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_outgoing_trace_headers_append_to_baggage_span_streaming(
    sentry_init,
    capture_items,
    httpx2_client,
    httpx2_mock,
):
    httpx2_mock.add_response()

    sentry_init(
        traces_sample_rate=1.0,
        integrations=[Httpx2Integration()],
        release="d08ebdb9309e1b004c6f52202de58a09c2268e42",
        trace_lifecycle="stream",
    )

    url = "http://example.com/"

    items = capture_items("span")

    with mock.patch("sentry_sdk.tracing_utils.Random.randrange", return_value=500000):
        with sentry_sdk.traces.start_span(name="test"):
            if asyncio.iscoroutinefunction(httpx2_client.get):
                response = asyncio.run(
                    httpx2_client.get(url, headers={"baGGage": "custom=data"})
                )
            else:
                response = httpx2_client.get(url, headers={"baGGage": "custom=data"})

    sentry_sdk.flush()

    http_span = _get_http_client_span(items)

    baggage = response.request.headers["baggage"]
    assert baggage.startswith("custom=data,")
    assert f"sentry-trace_id={http_span['trace_id']}" in baggage
    assert "sentry-sample_rand=0.500000" in baggage
    assert "sentry-sampled=true" in baggage


def test_outgoing_trace_headers_span_streaming_no_current_span(
    sentry_init, httpx2_mock
):
    """
    Even when there is no active span, trace propagation headers should still
    be attached to outgoing requests when span streaming is enabled.

    This is deliberately different from the legacy (transaction-based) approach,
    which does not propagate outside of a transaction (see
    ``test_do_not_propagate_outside_transaction``). The streamed approach
    propagates from the current scope's propagation context regardless of
    whether a span is active.
    """
    httpx2_mock.add_response()

    sentry_init(
        traces_sample_rate=1.0,
        trace_propagation_targets=[MATCH_ALL],
        integrations=[Httpx2Integration()],
        trace_lifecycle="stream",
    )

    url = "http://example.com/"

    httpx2_client = httpx2.Client()

    # No start_span / start_transaction -> get_current_span() is None
    assert sentry_sdk.traces.get_current_span() is None

    response = httpx2_client.get(url)

    assert response.status_code == 200

    # Trace is still propagated from the scope's propagation context
    request_headers = httpx2_mock.get_request().headers
    assert "sentry-trace" in request_headers
    assert "baggage" in request_headers

    # The propagated headers describe a single, coherent trace: the trace_id in
    # sentry-trace matches the one carried in baggage.
    trace_id = request_headers["sentry-trace"].split("-")[0]
    assert f"sentry-trace_id={trace_id}" in request_headers["baggage"]


def test_outgoing_trace_headers_span_streaming_no_current_span_async(
    sentry_init, httpx2_mock
):
    """
    The async client must match the sync client: trace propagation headers are
    attached to outgoing requests even when there is no active span and span
    streaming is enabled.
    """
    httpx2_mock.add_response()

    sentry_init(
        traces_sample_rate=1.0,
        trace_propagation_targets=[MATCH_ALL],
        integrations=[Httpx2Integration()],
        trace_lifecycle="stream",
    )

    url = "http://example.com/"

    httpx2_client = httpx2.AsyncClient()

    # No start_span / start_transaction -> get_current_span() is None
    assert sentry_sdk.traces.get_current_span() is None

    response = asyncio.run(httpx2_client.get(url))

    assert response.status_code == 200

    # Trace is still propagated from the scope's propagation context
    request_headers = httpx2_mock.get_request().headers
    assert "sentry-trace" in request_headers
    assert "baggage" in request_headers

    # The propagated headers describe a single, coherent trace: the trace_id in
    # sentry-trace matches the one carried in baggage.
    trace_id = request_headers["sentry-trace"].split("-")[0]
    assert f"sentry-trace_id={trace_id}" in request_headers["baggage"]


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_request_source_disabled_span_streaming(
    sentry_init, capture_items, httpx2_client, httpx2_mock
):
    httpx2_mock.add_response()

    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
        enable_http_request_source=False,
        http_request_source_threshold_ms=0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    url = "http://example.com/"

    with sentry_sdk.traces.start_span(name="test"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    sentry_sdk.flush()

    http_span = _get_http_client_span(items)

    assert "code.line.number" not in http_span["attributes"]
    assert SPANDATA.CODE_NAMESPACE not in http_span["attributes"]
    assert "code.file.path" not in http_span["attributes"]
    assert SPANDATA.CODE_FUNCTION not in http_span["attributes"]


@pytest.mark.parametrize("enable_http_request_source", [None, True])
@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_request_source_enabled_span_streaming(
    sentry_init,
    capture_items,
    enable_http_request_source,
    httpx2_client,
    httpx2_mock,
):
    httpx2_mock.add_response()

    sentry_options = {
        "integrations": [Httpx2Integration()],
        "traces_sample_rate": 1.0,
        "http_request_source_threshold_ms": 0,
        "trace_lifecycle": "stream",
    }
    if enable_http_request_source is not None:
        sentry_options["enable_http_request_source"] = enable_http_request_source

    sentry_init(**sentry_options)

    items = capture_items("span")

    url = "http://example.com/"

    with sentry_sdk.traces.start_span(name="test"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    sentry_sdk.flush()

    http_span = _get_http_client_span(items)

    assert "code.line.number" in http_span["attributes"]
    assert SPANDATA.CODE_NAMESPACE in http_span["attributes"]
    assert "code.file.path" in http_span["attributes"]
    assert SPANDATA.CODE_FUNCTION in http_span["attributes"]


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_request_source_span_streaming(
    sentry_init, capture_items, httpx2_client, httpx2_mock
):
    httpx2_mock.add_response()

    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    url = "http://example.com/"

    with sentry_sdk.traces.start_span(name="test"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    sentry_sdk.flush()

    http_span = _get_http_client_span(items)

    assert "code.line.number" in http_span["attributes"]
    assert SPANDATA.CODE_NAMESPACE in http_span["attributes"]
    assert "code.file.path" in http_span["attributes"]
    assert SPANDATA.CODE_FUNCTION in http_span["attributes"]

    assert type(http_span["attributes"]["code.line.number"]) == int
    assert http_span["attributes"]["code.line.number"] > 0
    assert (
        http_span["attributes"][SPANDATA.CODE_NAMESPACE]
        == "tests.integrations.httpx2.test_httpx2"
    )
    assert http_span["attributes"]["code.file.path"].endswith(
        "tests/integrations/httpx2/test_httpx2.py"
    )

    is_relative_path = http_span["attributes"]["code.file.path"][0] != os.sep
    assert is_relative_path

    assert (
        http_span["attributes"][SPANDATA.CODE_FUNCTION]
        == "test_request_source_span_streaming"
    )


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_request_source_with_module_in_search_path_span_streaming(
    sentry_init, capture_items, httpx2_client, httpx2_mock
):
    """
    Test that request source is relative to the path of the module it ran in
    """
    httpx2_mock.add_response()

    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        http_request_source_threshold_ms=0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    url = "http://example.com/"

    with sentry_sdk.traces.start_span(name="test"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            from httpx2_helpers.helpers import async_get_request_with_client

            asyncio.run(async_get_request_with_client(httpx2_client, url))
        else:
            from httpx2_helpers.helpers import get_request_with_client

            get_request_with_client(httpx2_client, url)

    sentry_sdk.flush()

    http_span = _get_http_client_span(items)

    assert "code.line.number" in http_span["attributes"]
    assert SPANDATA.CODE_NAMESPACE in http_span["attributes"]
    assert "code.file.path" in http_span["attributes"]
    assert SPANDATA.CODE_FUNCTION in http_span["attributes"]

    assert type(http_span["attributes"]["code.line.number"]) == int
    assert http_span["attributes"]["code.line.number"] > 0
    assert http_span["attributes"][SPANDATA.CODE_NAMESPACE] == "httpx2_helpers.helpers"
    assert http_span["attributes"]["code.file.path"] == "httpx2_helpers/helpers.py"

    is_relative_path = http_span["attributes"]["code.file.path"][0] != os.sep
    assert is_relative_path

    if asyncio.iscoroutinefunction(httpx2_client.get):
        assert (
            http_span["attributes"][SPANDATA.CODE_FUNCTION]
            == "async_get_request_with_client"
        )
    else:
        assert (
            http_span["attributes"][SPANDATA.CODE_FUNCTION] == "get_request_with_client"
        )


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_no_request_source_if_duration_too_short_span_streaming(
    sentry_init, capture_items, httpx2_client, httpx2_mock
):
    httpx2_mock.add_response()

    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        # Threshold so high no real request will ever exceed it
        http_request_source_threshold_ms=9999999,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    url = "http://example.com/"

    with sentry_sdk.traces.start_span(name="test"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    sentry_sdk.flush()

    http_span = _get_http_client_span(items)

    assert "code.line.number" not in http_span["attributes"]
    assert SPANDATA.CODE_NAMESPACE not in http_span["attributes"]
    assert "code.file.path" not in http_span["attributes"]
    assert SPANDATA.CODE_FUNCTION not in http_span["attributes"]


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_request_source_if_duration_over_threshold_span_streaming(
    sentry_init, capture_items, httpx2_client, httpx2_mock
):
    httpx2_mock.add_response()

    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
        enable_http_request_source=True,
        # Threshold of 0 means any non-zero duration qualifies
        http_request_source_threshold_ms=0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    url = "http://example.com/"

    with sentry_sdk.traces.start_span(name="test"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    sentry_sdk.flush()

    http_span = _get_http_client_span(items)

    assert "code.line.number" in http_span["attributes"]
    assert SPANDATA.CODE_NAMESPACE in http_span["attributes"]
    assert "code.file.path" in http_span["attributes"]
    assert SPANDATA.CODE_FUNCTION in http_span["attributes"]

    assert type(http_span["attributes"]["code.line.number"]) == int
    assert http_span["attributes"]["code.line.number"] > 0
    assert (
        http_span["attributes"][SPANDATA.CODE_NAMESPACE]
        == "tests.integrations.httpx2.test_httpx2"
    )
    assert http_span["attributes"]["code.file.path"].endswith(
        "tests/integrations/httpx2/test_httpx2.py"
    )

    is_relative_path = http_span["attributes"]["code.file.path"][0] != os.sep
    assert is_relative_path

    assert (
        http_span["attributes"][SPANDATA.CODE_FUNCTION]
        == "test_request_source_if_duration_over_threshold_span_streaming"
    )


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_span_origin_span_streaming(
    sentry_init, capture_items, httpx2_client, httpx2_mock
):
    httpx2_mock.add_response()

    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    url = "http://example.com/"

    with sentry_sdk.traces.start_span(name="test"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    sentry_sdk.flush()

    http_span = _get_http_client_span(items)

    assert http_span["attributes"]["sentry.origin"] == "auto.http.httpx2"


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_http_url_attributes_span_streaming(
    sentry_init, capture_items, httpx2_client, httpx2_mock
):
    httpx2_mock.add_response()

    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
        send_default_pii=True,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    url = "http://example.com/?foo=bar#frag"

    with sentry_sdk.traces.start_span(name="test"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    sentry_sdk.flush()

    http_span = _get_http_client_span(items)

    assert http_span["attributes"]["http.request.method"] == "GET"
    assert http_span["attributes"]["url.full"] == "http://example.com/"
    assert http_span["attributes"]["url.query"] == "foo=bar"
    assert http_span["attributes"]["url.fragment"] == "frag"
    assert http_span["attributes"]["http.response.status_code"] == 200


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_http_url_attributes_no_query_or_fragment_span_streaming(
    sentry_init, capture_items, httpx2_client, httpx2_mock
):
    httpx2_mock.add_response()

    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
        send_default_pii=True,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    url = "http://example.com/"

    with sentry_sdk.traces.start_span(name="test"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    sentry_sdk.flush()

    http_span = _get_http_client_span(items)

    assert http_span["attributes"]["http.request.method"] == "GET"
    assert http_span["attributes"]["url.full"] == "http://example.com/"
    assert "url.query" not in http_span["attributes"]
    assert "url.fragment" not in http_span["attributes"]
    assert http_span["attributes"]["http.response.status_code"] == 200


@pytest.mark.parametrize(
    "httpx2_client",
    (httpx2.Client(), httpx2.AsyncClient()),
)
def test_http_url_attributes_pii_disabled_span_streaming(
    sentry_init, capture_items, httpx2_client, httpx2_mock
):
    httpx2_mock.add_response()

    sentry_init(
        integrations=[Httpx2Integration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    url = "http://example.com/?foo=bar#frag"

    with sentry_sdk.traces.start_span(name="test"):
        if asyncio.iscoroutinefunction(httpx2_client.get):
            asyncio.run(httpx2_client.get(url))
        else:
            httpx2_client.get(url)

    sentry_sdk.flush()

    http_span = _get_http_client_span(items)

    assert http_span["attributes"]["http.request.method"] == "GET"
    assert "url.full" not in http_span["attributes"]
    assert "url.query" not in http_span["attributes"]
    assert "url.fragment" not in http_span["attributes"]
    assert http_span["attributes"]["http.response.status_code"] == 200



import pytest

from sentry_sdk import get_client
from sentry_sdk.integrations import _INTEGRATION_DEACTIVATES

try:
    from sentry_sdk.integrations.langchain import LangchainIntegration

    has_langchain = True
except Exception:
    has_langchain = False

try:
    from sentry_sdk.integrations.openai import OpenAIIntegration

    has_openai = True
except Exception:
    has_openai = False

try:
    from sentry_sdk.integrations.anthropic import AnthropicIntegration

    has_anthropic = True
except Exception:
    has_anthropic = False


try:
    from sentry_sdk.integrations.openai_agents import OpenAIAgentsIntegration

    has_openai_agents = True
except Exception:
    has_openai_agents = False

try:
    from sentry_sdk.integrations.pydantic_ai import PydanticAIIntegration

    has_pydantic_ai = True
except Exception:
    has_pydantic_ai = False


pytestmark = pytest.mark.skipif(
    not (
        has_langchain
        and has_openai
        and has_anthropic
        and has_openai_agents
        and has_pydantic_ai
    ),
    reason="Requires langchain, openai, and anthropic packages to be installed",
)


def test_integration_deactivates_map_exists():
    assert "langchain" in _INTEGRATION_DEACTIVATES
    assert "openai" in _INTEGRATION_DEACTIVATES["langchain"]
    assert "anthropic" in _INTEGRATION_DEACTIVATES["langchain"]
    assert "google_genai" in _INTEGRATION_DEACTIVATES["langchain"]
    assert "openai_agents" in _INTEGRATION_DEACTIVATES
    assert "openai" in _INTEGRATION_DEACTIVATES["openai_agents"]
    assert "pydantic_ai" in _INTEGRATION_DEACTIVATES
    assert "openai" in _INTEGRATION_DEACTIVATES["pydantic_ai"]
    assert "anthropic" in _INTEGRATION_DEACTIVATES["pydantic_ai"]


def test_langchain_auto_deactivates_openai_and_anthropic(
    sentry_init, reset_integrations
):
    sentry_init(
        default_integrations=False,
        auto_enabling_integrations=True,
    )

    client = get_client()
    integration_types = {
        type(integration) for integration in client.integrations.values()
    }

    if LangchainIntegration in integration_types:
        assert OpenAIIntegration not in integration_types
        assert AnthropicIntegration not in integration_types


def test_user_can_override_with_explicit_openai(sentry_init, reset_integrations):
    sentry_init(
        default_integrations=False,
        auto_enabling_integrations=True,
        integrations=[OpenAIIntegration()],
    )

    client = get_client()
    integration_types = {
        type(integration) for integration in client.integrations.values()
    }

    assert OpenAIIntegration in integration_types


def test_user_can_override_with_explicit_anthropic(sentry_init, reset_integrations):
    sentry_init(
        default_integrations=False,
        auto_enabling_integrations=True,
        integrations=[AnthropicIntegration()],
    )

    client = get_client()
    integration_types = {
        type(integration) for integration in client.integrations.values()
    }

    assert AnthropicIntegration in integration_types


def test_user_can_override_with_both_explicit_integrations(
    sentry_init, reset_integrations
):
    sentry_init(
        default_integrations=False,
        auto_enabling_integrations=True,
        integrations=[OpenAIIntegration(), AnthropicIntegration()],
    )

    client = get_client()
    integration_types = {
        type(integration) for integration in client.integrations.values()
    }

    assert OpenAIIntegration in integration_types
    assert AnthropicIntegration in integration_types


def test_disabling_integrations_allows_openai_and_anthropic(
    sentry_init, reset_integrations
):
    sentry_init(
        default_integrations=False,
        auto_enabling_integrations=True,
        disabled_integrations=[
            LangchainIntegration,
            OpenAIAgentsIntegration,
            PydanticAIIntegration,
        ],
    )

    client = get_client()
    integration_types = {
        type(integration) for integration in client.integrations.values()
    }

    assert LangchainIntegration not in integration_types


def test_explicit_langchain_still_deactivates_others(sentry_init, reset_integrations):
    sentry_init(
        default_integrations=False,
        auto_enabling_integrations=False,
        integrations=[LangchainIntegration()],
    )

    client = get_client()
    integration_types = {
        type(integration) for integration in client.integrations.values()
    }

    if LangchainIntegration in integration_types:
        assert OpenAIIntegration not in integration_types
        assert AnthropicIntegration not in integration_types


def test_langchain_and_openai_both_explicit_both_active(
    sentry_init, reset_integrations
):
    sentry_init(
        default_integrations=False,
        auto_enabling_integrations=False,
        integrations=[LangchainIntegration(), OpenAIIntegration()],
    )

    client = get_client()
    integration_types = {
        type(integration) for integration in client.integrations.values()
    }

    assert LangchainIntegration in integration_types
    assert OpenAIIntegration in integration_types


def test_no_langchain_means_openai_and_anthropic_can_auto_enable(
    sentry_init, reset_integrations, monkeypatch
):
    import sentry_sdk.integrations

    old_iter = sentry_sdk.integrations.iter_default_integrations

    def filtered_iter(with_auto_enabling):
        for cls in old_iter(with_auto_enabling):
            if cls.identifier != "langchain":
                yield cls

    monkeypatch.setattr(
        sentry_sdk.integrations, "iter_default_integrations", filtered_iter
    )

    sentry_init(
        default_integrations=False,
        auto_enabling_integrations=True,
    )

    client = get_client()
    integration_types = {
        type(integration) for integration in client.integrations.values()
    }

    assert LangchainIntegration not in integration_types


def test_deactivation_with_default_integrations_enabled(
    sentry_init, reset_integrations
):
    sentry_init(
        default_integrations=True,
        auto_enabling_integrations=True,
    )

    client = get_client()
    integration_types = {
        type(integration) for integration in client.integrations.values()
    }

    if LangchainIntegration in integration_types:
        assert OpenAIIntegration not in integration_types
        assert AnthropicIntegration not in integration_types


def test_only_auto_enabling_integrations_without_defaults(
    sentry_init, reset_integrations
):
    sentry_init(
        default_integrations=False,
        auto_enabling_integrations=True,
    )

    client = get_client()
    integration_types = {
        type(integration) for integration in client.integrations.values()
    }

    if LangchainIntegration in integration_types:
        assert OpenAIIntegration not in integration_types
        assert AnthropicIntegration not in integration_types



from unittest import mock

import pytest
import rq
from fakeredis import FakeStrictRedis

import sentry_sdk
from sentry_sdk import start_transaction
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations.rq import RqIntegration
from sentry_sdk.utils import SENSITIVE_DATA_SUBSTITUTE, parse_version


@pytest.fixture(autouse=True)
def _patch_rq_get_server_version(monkeypatch):
    """
    Patch RQ lower than 1.5.1 to work with fakeredis.

    https://github.com/jamesls/fakeredis/issues/273
    """
    try:
        from distutils.version import StrictVersion
    except ImportError:
        return

    if parse_version(rq.VERSION) <= (1, 5, 1):
        for k in (
            "rq.job.Job.get_redis_server_version",
            "rq.worker.Worker.get_redis_server_version",
        ):
            try:
                monkeypatch.setattr(k, lambda _: StrictVersion("4.0.0"))
            except AttributeError:
                # old RQ Job/Worker doesn't have a get_redis_server_version attr
                pass


def crashing_job(foo):
    1 / 0


def chew_up_shoes(dog, human, shoes):
    raise Exception("{}!! Why did you eat {}'s {}??".format(dog, human, shoes))


def do_trick(dog, trick):
    return "{}, can you {}? Good dog!".format(dog, trick)


@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("span_streaming", [True, False])
def test_basic(
    sentry_init,
    capture_events,
    capture_items,
    send_default_pii,
    span_streaming,
):
    sentry_init(
        integrations=[RqIntegration()],
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    queue = rq.Queue(connection=FakeStrictRedis())
    worker = rq.SimpleWorker([queue], connection=queue.connection)

    if span_streaming:
        items = capture_items("event")

        queue.enqueue(crashing_job, foo=42)
        worker.work(burst=True)

        (event,) = (item.payload for item in items)

        (exception,) = event["exception"]["values"]
        assert exception["type"] == "ZeroDivisionError"
        assert exception["mechanism"]["type"] == "rq"
        assert exception["stacktrace"]["frames"][-1]["vars"]["foo"] == "42"
    else:
        events = capture_events()

        queue.enqueue(crashing_job, foo=42)
        worker.work(burst=True)

        (event,) = events

        (exception,) = event["exception"]["values"]
        assert exception["type"] == "ZeroDivisionError"
        assert exception["mechanism"]["type"] == "rq"
        assert exception["stacktrace"]["frames"][-1]["vars"]["foo"] == "42"
        assert event["transaction"] == "tests.integrations.rq.test_rq.crashing_job"

    extra = event["extra"]["rq-job"]
    if send_default_pii:
        assert extra["args"] == []
        assert extra["kwargs"] == {"foo": 42}
    else:
        assert extra["args"] == SENSITIVE_DATA_SUBSTITUTE
        assert extra["kwargs"] == SENSITIVE_DATA_SUBSTITUTE
    assert extra["description"] == "tests.integrations.rq.test_rq.crashing_job(foo=42)"
    assert extra["func"] == "tests.integrations.rq.test_rq.crashing_job"
    assert "job_id" in extra
    assert "enqueued_at" in extra

    # older versions don't persist started_at correctly
    if tuple(map(int, rq.VERSION.split("."))) >= (0, 9):
        assert "started_at" in extra


@pytest.mark.parametrize("span_streaming", [True, False])
def test_transport_shutdown(
    sentry_init, capture_events_forksafe, capture_items_forksafe, span_streaming
):
    sentry_init(
        integrations=[RqIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    queue = rq.Queue(connection=FakeStrictRedis())
    worker = rq.Worker([queue], connection=queue.connection)

    if span_streaming:
        items = capture_items_forksafe("event")

        queue.enqueue(crashing_job, foo=42)
        worker.work(burst=True)

        captured_items = items.read_event()
        items.read_flush()

        event = next(item["payload"] for item in captured_items)
        (exception,) = event["exception"]["values"]
        assert exception["type"] == "ZeroDivisionError"
    else:
        events = capture_events_forksafe()

        queue.enqueue(crashing_job, foo=42)
        worker.work(burst=True)

        event = events.read_event()
        events.read_flush()

        (exception,) = event["exception"]["values"]
        assert exception["type"] == "ZeroDivisionError"


@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("span_streaming", [True, False])
def test_transaction_with_error(
    sentry_init,
    capture_events,
    capture_items,
    DictionaryContaining,
    send_default_pii,
    span_streaming,
):
    sentry_init(
        integrations=[RqIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    queue = rq.Queue(connection=FakeStrictRedis())
    worker = rq.SimpleWorker([queue], connection=queue.connection)

    if span_streaming:
        items = capture_items("event", "span")

        queue.enqueue(chew_up_shoes, "Charlie", "Katie", shoes="flip-flops")
        worker.work(burst=True)

        (error_event,) = (item.payload for item in items if item.type == "event")

        assert (
            error_event["transaction"] == "tests.integrations.rq.test_rq.chew_up_shoes"
        )
        assert error_event["contexts"]["trace"]["op"] == "queue.task.rq"
        assert error_event["exception"]["values"][0]["type"] == "Exception"
        assert (
            error_event["exception"]["values"][0]["value"]
            == "Charlie!! Why did you eat Katie's flip-flops??"
        )

        sentry_sdk.flush()
        spans = [item.payload for item in items if item.type == "span"]
        (span,) = (
            span
            for span in spans
            if span["attributes"].get("sentry.op") == "queue.task.rq"
        )

        assert span["trace_id"] == error_event["contexts"]["trace"]["trace_id"]
        assert span["span_id"] == error_event["contexts"]["trace"]["span_id"]
        assert span["attributes"]["sentry.op"] == error_event["contexts"]["trace"]["op"]
        assert (
            span["attributes"]["sentry.origin"]
            == error_event["contexts"]["trace"]["origin"]
        )
        assert (
            span["attributes"][SPANDATA.CODE_FUNCTION_NAME]
            == "tests.integrations.rq.test_rq.chew_up_shoes"
        )
    else:
        events = capture_events()

        queue.enqueue(chew_up_shoes, "Charlie", "Katie", shoes="flip-flops")
        worker.work(burst=True)

        error_event, envelope = events

        assert (
            error_event["transaction"] == "tests.integrations.rq.test_rq.chew_up_shoes"
        )
        assert error_event["contexts"]["trace"]["op"] == "queue.task.rq"
        assert error_event["exception"]["values"][0]["type"] == "Exception"
        assert (
            error_event["exception"]["values"][0]["value"]
            == "Charlie!! Why did you eat Katie's flip-flops??"
        )

        assert envelope["type"] == "transaction"
        assert envelope["contexts"]["trace"] == error_event["contexts"]["trace"]
        assert envelope["transaction"] == error_event["transaction"]
        assert envelope["extra"]["rq-job"] == DictionaryContaining(
            {
                "args": (
                    ["Charlie", "Katie"]
                    if send_default_pii
                    else SENSITIVE_DATA_SUBSTITUTE
                ),
                "kwargs": (
                    {"shoes": "flip-flops"}
                    if send_default_pii
                    else SENSITIVE_DATA_SUBSTITUTE
                ),
                "func": "tests.integrations.rq.test_rq.chew_up_shoes",
                "description": "tests.integrations.rq.test_rq.chew_up_shoes('Charlie', 'Katie', shoes='flip-flops')",
            }
        )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_error_has_trace_context_if_tracing_disabled(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[RqIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    queue = rq.Queue(connection=FakeStrictRedis())
    worker = rq.SimpleWorker([queue], connection=queue.connection)

    if span_streaming:
        items = capture_items("event")

        queue.enqueue(crashing_job, foo=None)
        worker.work(burst=True)

        (error_event,) = (item.payload for item in items)
    else:
        events = capture_events()

        queue.enqueue(crashing_job, foo=None)
        worker.work(burst=True)

        (error_event,) = events

    assert error_event["contexts"]["trace"]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_tracing_enabled(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[RqIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    queue = rq.Queue(connection=FakeStrictRedis())
    worker = rq.SimpleWorker([queue], connection=queue.connection)

    if span_streaming:
        items = capture_items("event", "span")

        with sentry_sdk.traces.start_span(
            name="custom parent",
            attributes={
                "sentry.op": "rq transaction",
            },
        ) as span:
            queue.enqueue(crashing_job, foo=None)
            worker.work(burst=True)

        (error_event,) = (item.payload for item in items if item.type == "event")

        assert error_event["contexts"]["trace"]["trace_id"] == span.trace_id

        sentry_sdk.flush()
        spans = [item.payload for item in items if item.type == "span"]
        (span,) = (
            span
            for span in spans
            if span["attributes"].get("sentry.op") == "queue.task.rq"
        )

        assert span["trace_id"] == error_event["contexts"]["trace"]["trace_id"]
        assert span["span_id"] == error_event["contexts"]["trace"]["span_id"]
        assert span["attributes"]["sentry.op"] == error_event["contexts"]["trace"]["op"]
        assert (
            span["attributes"]["sentry.origin"]
            == error_event["contexts"]["trace"]["origin"]
        )
    else:
        events = capture_events()

        with start_transaction(op="rq transaction") as transaction:
            queue.enqueue(crashing_job, foo=None)
            worker.work(burst=True)

        error_event, envelope, _ = events

        assert (
            error_event["transaction"] == "tests.integrations.rq.test_rq.crashing_job"
        )
        assert error_event["contexts"]["trace"]["trace_id"] == transaction.trace_id

        assert envelope["contexts"]["trace"] == error_event["contexts"]["trace"]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_tracing_disabled(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[RqIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    queue = rq.Queue(connection=FakeStrictRedis())
    worker = rq.SimpleWorker([queue], connection=queue.connection)

    if span_streaming:
        items = capture_items("event")

        queue.enqueue(crashing_job, foo=None)
        worker.work(burst=True)

        (error_event,) = (item.payload for item in items)

        assert error_event["contexts"]["trace"]["trace_id"]
    else:
        events = capture_events()

        scope = sentry_sdk.get_isolation_scope()
        queue.enqueue(crashing_job, foo=None)
        worker.work(burst=True)

        (error_event,) = events

        assert (
            error_event["transaction"] == "tests.integrations.rq.test_rq.crashing_job"
        )
        assert (
            error_event["contexts"]["trace"]["trace_id"]
            == scope._propagation_context.trace_id
        )


@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("span_streaming", [True, False])
def test_transaction_no_error(
    sentry_init,
    capture_events,
    capture_items,
    DictionaryContaining,
    send_default_pii,
    span_streaming,
):
    sentry_init(
        integrations=[RqIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    queue = rq.Queue(connection=FakeStrictRedis())
    worker = rq.SimpleWorker([queue], connection=queue.connection)

    if span_streaming:
        items = capture_items("span")

        queue.enqueue(do_trick, "Maisey", trick="kangaroo")
        worker.work(burst=True)

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        (span,) = (
            span
            for span in spans
            if span["attributes"].get("sentry.op") == "queue.task.rq"
        )

        assert span["attributes"]["sentry.op"] == "queue.task.rq"
        assert span["name"] == "tests.integrations.rq.test_rq.do_trick"
        assert span["attributes"][SPANDATA.MESSAGING_DESTINATION_NAME] == queue.name
    else:
        events = capture_events()

        queue.enqueue(do_trick, "Maisey", trick="kangaroo")
        worker.work(burst=True)

        envelope = events[0]

        assert envelope["type"] == "transaction"
        assert envelope["contexts"]["trace"]["op"] == "queue.task.rq"
        assert envelope["transaction"] == "tests.integrations.rq.test_rq.do_trick"
        assert (
            envelope["contexts"]["trace"]["data"][SPANDATA.MESSAGING_DESTINATION_NAME]
            == queue.name
        )
        assert envelope["extra"]["rq-job"] == DictionaryContaining(
            {
                "args": ["Maisey"] if send_default_pii else SENSITIVE_DATA_SUBSTITUTE,
                "kwargs": (
                    {"trick": "kangaroo"}
                    if send_default_pii
                    else SENSITIVE_DATA_SUBSTITUTE
                ),
                "func": "tests.integrations.rq.test_rq.do_trick",
                "description": "tests.integrations.rq.test_rq.do_trick('Maisey', trick='kangaroo')",
            }
        )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_traces_sampler_gets_correct_values_in_sampling_context(
    sentry_init,
    DictionaryContaining,
    ObjectDescribedBy,
    span_streaming,  # noqa:N803
):
    traces_sampler = mock.Mock(return_value=True)
    sentry_init(
        integrations=[RqIntegration()],
        traces_sampler=traces_sampler,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    queue = rq.Queue(connection=FakeStrictRedis())
    worker = rq.SimpleWorker([queue], connection=queue.connection)

    queue.enqueue(do_trick, "Bodhi", trick="roll over")
    worker.work(burst=True)

    traces_sampler.assert_any_call(
        DictionaryContaining(
            {
                "rq_job": ObjectDescribedBy(
                    type=rq.job.Job,
                    attrs={
                        "description": "tests.integrations.rq.test_rq.do_trick('Bodhi', trick='roll over')",
                        "result": "Bodhi, can you roll over? Good dog!",
                        "func_name": "tests.integrations.rq.test_rq.do_trick",
                        "args": ("Bodhi",),
                        "kwargs": {"trick": "roll over"},
                    },
                ),
            }
        )
    )


@pytest.mark.skipif(
    parse_version(rq.__version__) < (1, 5), reason="At least rq-1.5 required"
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_job_with_retries(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[RqIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    queue = rq.Queue(connection=FakeStrictRedis())
    worker = rq.SimpleWorker([queue], connection=queue.connection)

    if span_streaming:
        items = capture_items("event")

        queue.enqueue(crashing_job, foo=42, retry=rq.Retry(max=1))
        worker.work(burst=True)
        events = [item.payload for item in items]
    else:
        events = capture_events()

        queue.enqueue(crashing_job, foo=42, retry=rq.Retry(max=1))
        worker.work(burst=True)

    assert len(events) == 1


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[RqIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    queue = rq.Queue(connection=FakeStrictRedis())
    worker = rq.SimpleWorker([queue], connection=queue.connection)

    if span_streaming:
        items = capture_items("span")

        queue.enqueue(do_trick, "Maisey", trick="kangaroo")
        worker.work(burst=True)

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        (span,) = (
            span
            for span in spans
            if span["attributes"].get("sentry.op") == "queue.task.rq"
        )

        assert span["attributes"]["sentry.origin"] == "auto.queue.rq"
    else:
        events = capture_events()

        queue.enqueue(do_trick, "Maisey", trick="kangaroo")
        worker.work(burst=True)

        (event,) = events

        assert event["contexts"]["trace"]["origin"] == "auto.queue.rq"



"""
This tests test for the correctness of the dynamic sampling context (DSC) in the trace header of envelopes.

The DSC is defined here:
https://develop.sentry.dev/sdk/telemetry/traces/dynamic-sampling-context/#dsc-specification

The DSC is propagated between service using a header called "baggage".
This is not tested in this file.
"""

from unittest import mock

import pytest

import sentry_sdk
from tests.conftest import TestTransportWithOptions


def test_dsc_head_of_trace(sentry_init, capture_envelopes):
    """
    Our service is the head of the trace (it starts a new trace)
    and sends a transaction event to Sentry.
    """
    sentry_init(
        dsn="https://mysecret@o1234.ingest.sentry.io/12312012",
        release="myapp@0.0.1",
        environment="canary",
        traces_sample_rate=1.0,
        transport=TestTransportWithOptions,
    )
    envelopes = capture_envelopes()

    # We start a new transaction
    with sentry_sdk.start_transaction(name="foo"):
        pass

    assert len(envelopes) == 1

    transaction_envelope = envelopes[0]
    envelope_trace_header = transaction_envelope.headers["trace"]

    assert "trace_id" in envelope_trace_header
    assert type(envelope_trace_header["trace_id"]) == str

    assert "public_key" in envelope_trace_header
    assert type(envelope_trace_header["public_key"]) == str
    assert envelope_trace_header["public_key"] == "mysecret"

    assert "org_id" in envelope_trace_header
    assert type(envelope_trace_header["org_id"]) == str
    assert envelope_trace_header["org_id"] == "1234"

    assert "sample_rate" in envelope_trace_header
    assert type(envelope_trace_header["sample_rate"]) == str
    assert envelope_trace_header["sample_rate"] == "1.0"

    assert "sampled" in envelope_trace_header
    assert type(envelope_trace_header["sampled"]) == str
    assert envelope_trace_header["sampled"] == "true"

    assert "release" in envelope_trace_header
    assert type(envelope_trace_header["release"]) == str
    assert envelope_trace_header["release"] == "myapp@0.0.1"

    assert "environment" in envelope_trace_header
    assert type(envelope_trace_header["environment"]) == str
    assert envelope_trace_header["environment"] == "canary"

    assert "transaction" in envelope_trace_header
    assert type(envelope_trace_header["transaction"]) == str
    assert envelope_trace_header["transaction"] == "foo"


def test_dsc_head_of_trace_uses_custom_org_id(sentry_init, capture_envelopes):
    """
    Our service is the head of the trace (it starts a new trace)
    and sends a transaction event to Sentry.
    """
    sentry_init(
        dsn="https://mysecret@o1234.ingest.sentry.io/12312012",
        org_id="9999",
        release="myapp@0.0.1",
        environment="canary",
        traces_sample_rate=1.0,
        transport=TestTransportWithOptions,
    )
    envelopes = capture_envelopes()

    # We start a new transaction
    with sentry_sdk.start_transaction(name="foo"):
        pass

    assert len(envelopes) == 1

    transaction_envelope = envelopes[0]
    envelope_trace_header = transaction_envelope.headers["trace"]

    assert "org_id" in envelope_trace_header
    assert type(envelope_trace_header["org_id"]) == str
    assert envelope_trace_header["org_id"] == "9999"


def test_dsc_continuation_of_trace(sentry_init, capture_envelopes):
    """
    Another service calls our service and passes tracing information to us.
    Our service is continuing the trace and sends a transaction event to Sentry.
    """
    sentry_init(
        dsn="https://mysecret@o1234.ingest.sentry.io/12312012",
        release="myapp@0.0.1",
        environment="canary",
        traces_sample_rate=1.0,
        transport=TestTransportWithOptions,
    )
    envelopes = capture_envelopes()

    # This is what the upstream service sends us
    sentry_trace = "771a43a4192642f0b136d5159a501700-1234567890abcdef-1"
    baggage = (
        "other-vendor-value-1=foo;bar;baz, "
        "sentry-trace_id=771a43a4192642f0b136d5159a501700, "
        "sentry-public_key=frontendpublickey, "
        "sentry-sample_rate=0.01337, "
        "sentry-sampled=true, "
        "sentry-release=myfrontend@1.2.3, "
        "sentry-environment=bird, "
        "sentry-transaction=bar, "
        "other-vendor-value-2=foo;bar;"
    )
    incoming_http_headers = {
        "HTTP_SENTRY_TRACE": sentry_trace,
        "HTTP_BAGGAGE": baggage,
    }

    # We continue the incoming trace and start a new transaction
    transaction = sentry_sdk.continue_trace(incoming_http_headers)
    with sentry_sdk.start_transaction(transaction, name="foo"):
        pass

    assert len(envelopes) == 1

    transaction_envelope = envelopes[0]
    envelope_trace_header = transaction_envelope.headers["trace"]

    assert "trace_id" in envelope_trace_header
    assert type(envelope_trace_header["trace_id"]) == str
    assert envelope_trace_header["trace_id"] == "771a43a4192642f0b136d5159a501700"

    assert "public_key" in envelope_trace_header
    assert type(envelope_trace_header["public_key"]) == str
    assert envelope_trace_header["public_key"] == "frontendpublickey"

    assert "sample_rate" in envelope_trace_header
    assert type(envelope_trace_header["sample_rate"]) == str
    assert envelope_trace_header["sample_rate"] == "1.0"

    assert "sampled" in envelope_trace_header
    assert type(envelope_trace_header["sampled"]) == str
    assert envelope_trace_header["sampled"] == "true"

    assert "release" in envelope_trace_header
    assert type(envelope_trace_header["release"]) == str
    assert envelope_trace_header["release"] == "myfrontend@1.2.3"

    assert "environment" in envelope_trace_header
    assert type(envelope_trace_header["environment"]) == str
    assert envelope_trace_header["environment"] == "bird"

    assert "transaction" in envelope_trace_header
    assert type(envelope_trace_header["transaction"]) == str
    assert envelope_trace_header["transaction"] == "bar"


def test_dsc_continuation_of_trace_sample_rate_changed_in_traces_sampler(
    sentry_init, capture_envelopes
):
    """
    Another service calls our service and passes tracing information to us.
    Our service is continuing the trace, but modifies the sample rate.
    The DSC propagated further should contain the updated sample rate.
    """

    def my_traces_sampler(sampling_context):
        return 0.25

    sentry_init(
        dsn="https://mysecret@o1234.ingest.sentry.io/12312012",
        release="myapp@0.0.1",
        environment="canary",
        traces_sampler=my_traces_sampler,
        transport=TestTransportWithOptions,
    )
    envelopes = capture_envelopes()

    # This is what the upstream service sends us
    sentry_trace = "771a43a4192642f0b136d5159a501700-1234567890abcdef-1"
    baggage = (
        "other-vendor-value-1=foo;bar;baz, "
        "sentry-trace_id=771a43a4192642f0b136d5159a501700, "
        "sentry-public_key=frontendpublickey, "
        "sentry-sample_rate=1.0, "
        "sentry-sampled=true, "
        "sentry-release=myfrontend@1.2.3, "
        "sentry-environment=bird, "
        "sentry-transaction=bar, "
        "other-vendor-value-2=foo;bar;"
    )
    incoming_http_headers = {
        "HTTP_SENTRY_TRACE": sentry_trace,
        "HTTP_BAGGAGE": baggage,
    }

    # We continue the incoming trace and start a new transaction
    with mock.patch("sentry_sdk.tracing_utils.Random.randrange", return_value=125000):
        transaction = sentry_sdk.continue_trace(incoming_http_headers)
        with sentry_sdk.start_transaction(transaction, name="foo"):
            pass

    assert len(envelopes) == 1

    transaction_envelope = envelopes[0]
    envelope_trace_header = transaction_envelope.headers["trace"]

    assert "trace_id" in envelope_trace_header
    assert type(envelope_trace_header["trace_id"]) == str
    assert envelope_trace_header["trace_id"] == "771a43a4192642f0b136d5159a501700"

    assert "public_key" in envelope_trace_header
    assert type(envelope_trace_header["public_key"]) == str
    assert envelope_trace_header["public_key"] == "frontendpublickey"

    assert "sample_rate" in envelope_trace_header
    assert type(envelope_trace_header["sample_rate"]) == str
    assert envelope_trace_header["sample_rate"] == "0.25"

    assert "sampled" in envelope_trace_header
    assert type(envelope_trace_header["sampled"]) == str
    assert envelope_trace_header["sampled"] == "true"

    assert "release" in envelope_trace_header
    assert type(envelope_trace_header["release"]) == str
    assert envelope_trace_header["release"] == "myfrontend@1.2.3"

    assert "environment" in envelope_trace_header
    assert type(envelope_trace_header["environment"]) == str
    assert envelope_trace_header["environment"] == "bird"

    assert "transaction" in envelope_trace_header
    assert type(envelope_trace_header["transaction"]) == str
    assert envelope_trace_header["transaction"] == "bar"


def test_dsc_issue(sentry_init, capture_envelopes):
    """
    Our service is a standalone service that does not have tracing enabled. Just uses Sentry for error reporting.
    """
    sentry_init(
        dsn="https://mysecret@o1234.ingest.sentry.io/12312012",
        release="myapp@0.0.1",
        environment="canary",
        transport=TestTransportWithOptions,
    )
    envelopes = capture_envelopes()

    # No transaction is started, just an error is captured
    try:
        1 / 0
    except ZeroDivisionError as exp:
        sentry_sdk.capture_exception(exp)

    assert len(envelopes) == 1

    error_envelope = envelopes[0]

    envelope_trace_header = error_envelope.headers["trace"]

    assert "trace_id" in envelope_trace_header
    assert type(envelope_trace_header["trace_id"]) == str

    assert "public_key" in envelope_trace_header
    assert type(envelope_trace_header["public_key"]) == str
    assert envelope_trace_header["public_key"] == "mysecret"

    assert "org_id" in envelope_trace_header
    assert type(envelope_trace_header["org_id"]) == str
    assert envelope_trace_header["org_id"] == "1234"

    assert "sample_rate" not in envelope_trace_header

    assert "sampled" not in envelope_trace_header

    assert "release" in envelope_trace_header
    assert type(envelope_trace_header["release"]) == str
    assert envelope_trace_header["release"] == "myapp@0.0.1"

    assert "environment" in envelope_trace_header
    assert type(envelope_trace_header["environment"]) == str
    assert envelope_trace_header["environment"] == "canary"

    assert "transaction" not in envelope_trace_header


def test_dsc_issue_with_tracing(sentry_init, capture_envelopes):
    """
    Our service has tracing enabled and an error occurs in an transaction.
    Envelopes containing errors also have the same DSC than the transaction envelopes.
    """
    sentry_init(
        dsn="https://mysecret@o1234.ingest.sentry.io/12312012",
        release="myapp@0.0.1",
        environment="canary",
        traces_sample_rate=1.0,
        transport=TestTransportWithOptions,
    )
    envelopes = capture_envelopes()

    # We start a new transaction and an error occurs
    with sentry_sdk.start_transaction(name="foo"):
        try:
            1 / 0
        except ZeroDivisionError as exp:
            sentry_sdk.capture_exception(exp)

    assert len(envelopes) == 2

    error_envelope, transaction_envelope = envelopes

    assert error_envelope.headers["trace"] == transaction_envelope.headers["trace"]

    envelope_trace_header = error_envelope.headers["trace"]

    assert "trace_id" in envelope_trace_header
    assert type(envelope_trace_header["trace_id"]) == str

    assert "public_key" in envelope_trace_header
    assert type(envelope_trace_header["public_key"]) == str
    assert envelope_trace_header["public_key"] == "mysecret"

    assert "org_id" in envelope_trace_header
    assert type(envelope_trace_header["org_id"]) == str
    assert envelope_trace_header["org_id"] == "1234"

    assert "sample_rate" in envelope_trace_header
    assert envelope_trace_header["sample_rate"] == "1.0"
    assert type(envelope_trace_header["sample_rate"]) == str

    assert "sampled" in envelope_trace_header
    assert type(envelope_trace_header["sampled"]) == str
    assert envelope_trace_header["sampled"] == "true"

    assert "release" in envelope_trace_header
    assert type(envelope_trace_header["release"]) == str
    assert envelope_trace_header["release"] == "myapp@0.0.1"

    assert "environment" in envelope_trace_header
    assert type(envelope_trace_header["environment"]) == str
    assert envelope_trace_header["environment"] == "canary"

    assert "transaction" in envelope_trace_header
    assert type(envelope_trace_header["transaction"]) == str
    assert envelope_trace_header["transaction"] == "foo"


@pytest.mark.parametrize(
    "traces_sample_rate",
    [
        0,  # no traces will be started, but if incoming traces will be continued (by our instrumentations, not happening in this test)
        None,  # no tracing at all. This service will never create transactions.
    ],
)
def test_dsc_issue_twp(sentry_init, capture_envelopes, traces_sample_rate):
    """
    Our service does not have tracing enabled, but we receive tracing information from an upstream service.
    Error envelopes still contain a DCS. This is called "tracing without performance" or TWP for short.

    This way if I have three services A, B, and C, and A and C have tracing enabled, but B does not,
    we still can see the full trace in Sentry, and associate errors send by service B to Sentry.
    (This test would be service B in this scenario)
    """
    sentry_init(
        dsn="https://mysecret@o1234.ingest.sentry.io/12312012",
        release="myapp@0.0.1",
        environment="canary",
        traces_sample_rate=traces_sample_rate,
        transport=TestTransportWithOptions,
    )
    envelopes = capture_envelopes()

    # This is what the upstream service sends us
    sentry_trace = "771a43a4192642f0b136d5159a501700-1234567890abcdef-1"
    baggage = (
        "other-vendor-value-1=foo;bar;baz, "
        "sentry-trace_id=771a43a4192642f0b136d5159a501700, "
        "sentry-public_key=frontendpublickey, "
        "sentry-sample_rate=0.01337, "
        "sentry-sampled=true, "
        "sentry-release=myfrontend@1.2.3, "
        "sentry-environment=bird, "
        "sentry-transaction=bar, "
        "other-vendor-value-2=foo;bar;"
    )
    incoming_http_headers = {
        "HTTP_SENTRY_TRACE": sentry_trace,
        "HTTP_BAGGAGE": baggage,
    }

    # We continue the trace (meaning: saving the incoming trace information on the scope)
    # but in this test, we do not start a transaction.
    sentry_sdk.continue_trace(incoming_http_headers)

    # No transaction is started, just an error is captured
    try:
        1 / 0
    except ZeroDivisionError as exp:
        sentry_sdk.capture_exception(exp)

    assert len(envelopes) == 1

    error_envelope = envelopes[0]

    envelope_trace_header = error_envelope.headers["trace"]

    assert "trace_id" in envelope_trace_header
    assert type(envelope_trace_header["trace_id"]) == str
    assert envelope_trace_header["trace_id"] == "771a43a4192642f0b136d5159a501700"

    assert "public_key" in envelope_trace_header
    assert type(envelope_trace_header["public_key"]) == str
    assert envelope_trace_header["public_key"] == "frontendpublickey"

    assert "sample_rate" in envelope_trace_header
    assert type(envelope_trace_header["sample_rate"]) == str
    assert envelope_trace_header["sample_rate"] == "0.01337"

    assert "sampled" in envelope_trace_header
    assert type(envelope_trace_header["sampled"]) == str
    assert envelope_trace_header["sampled"] == "true"

    assert "release" in envelope_trace_header
    assert type(envelope_trace_header["release"]) == str
    assert envelope_trace_header["release"] == "myfrontend@1.2.3"

    assert "environment" in envelope_trace_header
    assert type(envelope_trace_header["environment"]) == str
    assert envelope_trace_header["environment"] == "bird"

    assert "transaction" in envelope_trace_header
    assert type(envelope_trace_header["transaction"]) == str
    assert envelope_trace_header["transaction"] == "bar"
